from __future__ import annotations

import functools
import logging
import os
import time
from collections.abc import Iterator
from decimal import Decimal
from pathlib import Path

import psycopg
from pgtoolkit import ctl
from psycopg.conninfo import conninfo_to_dict
from psycopg.rows import args_row
from tenacity import retry
from tenacity.retry import retry_if_exception_type
from tenacity.stop import stop_after_attempt

from .. import cmd, db, exceptions
from ..models.system import BaseInstance, PostgreSQLInstance
from ..settings._postgresql import PostgreSQLVersion
from ..types import Status
from .models import WALSenderState

logger = logging.getLogger(__name__)


@functools.lru_cache(maxsize=len(PostgreSQLVersion) + 1)
def pg_ctl(bindir: Path) -> ctl.PGCtl:
    return ctl.PGCtl(bindir, run_command=cmd.run)


def is_ready(instance: PostgreSQLInstance) -> bool:
    """Return True if the instance is ready per pg_isready."""
    logger.debug("checking if PostgreSQL instance %s is ready", instance)
    pg_isready = str(instance.bindir / "pg_isready")
    user = instance._settings.postgresql.surole.name
    dsn = db.dsn(instance, user=user)
    env = libpq_environ(instance, user)
    r = cmd.run([pg_isready, "-d", dsn], env=env)
    if r.returncode == 0:
        return True
    assert r.returncode in (
        1,
        2,
    ), f"Unexpected exit status from pg_isready {r.returncode}: {r.stdout}, {r.stderr}"
    return False


def wait_ready(instance: PostgreSQLInstance, *, timeout: int = 10) -> None:
    for __ in range(timeout):
        if is_ready(instance):
            return
        time.sleep(1)
    raise exceptions.InstanceStateError(f"{instance} not ready after {timeout}s")


def status(instance: BaseInstance) -> Status:
    """Return the status of an instance."""
    logger.debug("get status of PostgreSQL instance %s", instance)
    # Map pg_ctl status (0, 3, 4) to our Status definition, assuming that 4
    # (data directory not specified) cannot happen at this point, otherwise
    # the 'instance' value would not exist.
    return Status(pg_ctl(instance.bindir).status(instance.datadir).value)


def is_running(instance: BaseInstance) -> bool:
    """Return True if the instance is running based on its status."""
    return status(instance) == Status.running


def check_status(instance: BaseInstance, expected: Status) -> None:
    """Check actual instance status with respected to `expected` one.

    :raises ~exceptions.InstanceStateError: in case the actual status is not expected.
    """
    if (st := status(instance)) != expected:
        raise exceptions.InstanceStateError(f"instance is {st.name}")


def get_data_checksums(instance: PostgreSQLInstance) -> bool:
    """Return True/False if data_checksums is enabled/disabled on instance."""
    controldata = pg_ctl(instance.bindir).controldata(instance.datadir)
    return controldata["Data page checksum version"] != "0"


def set_data_checksums(instance: PostgreSQLInstance, enabled: bool) -> None:
    """Enable/disable data checksums on instance."""
    if is_running(instance):
        raise exceptions.InstanceStateError(
            "could not alter data_checksums on a running instance"
        )
    action = "enable" if enabled else "disable"
    cmd.run(
        [
            str(instance.bindir / "pg_checksums"),
            f"--{action}",
            "--pgdata",
            str(instance.datadir),
        ],
        check=True,
    )


def logfile(
    instance: PostgreSQLInstance,
    *,
    timeout: float | None = None,
    poll_interval: float = 0.1,
) -> Iterator[Path]:
    """Yield the current log file by polling current_logfiles for changes.

    :raises ~exceptions.FileNotFoundError: if the current log file, matching
        first configured log_destination, is not found.
    :raises ~exceptions.SystemError: if the current log file cannot be opened
        for reading.
    :raises ValueError: if no record matching configured log_destination is
        found in current_logfiles (this indicates a misconfigured instance).
    :raises TimeoutError: if no new log file was polled from current_logfiles
        within specified 'timeout'.
    """
    config = instance.config()
    log_destination = config.get("log_destination", "stderr")
    assert isinstance(log_destination, str), log_destination
    log_destination = log_destination.split(",")[0]
    current_logfiles = instance.datadir / "current_logfiles"
    if not current_logfiles.exists():
        raise exceptions.FileNotFoundError(
            f"file 'current_logfiles' for instance {instance} not found"
        )

    @retry(
        retry=retry_if_exception_type(FileNotFoundError),
        stop=stop_after_attempt(2),
        reraise=True,
    )
    def logf() -> Path:
        """Get the current log file, matching configured log_destination.

        Retry in case the 'current_logfiles' file is unavailable for reading,
        which might happen as postgres typically re-creates it upon update.
        """
        with current_logfiles.open() as f:
            for line in f:
                destination, location = line.strip().split(None, maxsplit=1)
                if destination == log_destination:
                    break
            else:
                raise ValueError(
                    f"no record matching {log_destination!r} log destination found for instance {instance}"
                )
        fpath = Path(location)
        if not fpath.is_absolute():
            fpath = instance.datadir / fpath
        return fpath

    current_logfile = None
    start_time = time.monotonic()
    while True:
        f = logf()
        if f == current_logfile:
            if timeout is not None:
                if time.monotonic() - start_time >= timeout:
                    raise TimeoutError("timed out waiting for a new log file")
            time.sleep(poll_interval)
            continue
        current_logfile = f
        start_time = time.monotonic()
        yield current_logfile


def logs(
    instance: PostgreSQLInstance,
    *,
    timeout: float | None = None,
    poll_interval: float = 0.1,
) -> Iterator[str]:
    """Return the content of current log file as an iterator."""
    for fpath in logfile(instance, timeout=timeout, poll_interval=poll_interval):
        logger.info("reading logs of instance %s from %s", instance, fpath)
        try:
            with fpath.open() as f:
                yield from f
        except OSError as e:
            raise exceptions.SystemError(
                f"failed to read {fpath} on instance {instance}"
            ) from e


def replication_lag(instance: PostgreSQLInstance) -> Decimal | None:
    """Return the replication lag of a standby instance.

    The instance must be running; if the primary is not running, None is
    returned.

    :raises TypeError: if the instance is not a standby.
    """
    standby = instance.standby
    if standby is None:
        raise TypeError(f"{instance} is not a standby")

    try:
        with db.primary_connect(standby) as cnx:
            row = cnx.execute("SELECT pg_current_wal_lsn() AS lsn").fetchone()
    except psycopg.OperationalError as e:
        logger.warning("failed to connect to primary: %s", e)
        return None
    assert row is not None
    primary_lsn = row["lsn"]

    password = standby.password.get_secret_value() if standby.password else None
    dsn = db.dsn(
        instance,
        dbname="template1",
        user=instance._settings.postgresql.replrole,
        password=password,
    )
    with db.connect(dsn) as cnx:
        row = cnx.execute(
            "SELECT %s::pg_lsn - pg_last_wal_replay_lsn() AS lag", (primary_lsn,)
        ).fetchone()
    assert row is not None
    lag = row["lag"]
    assert isinstance(lag, Decimal)
    return lag


def wal_sender_state(instance: PostgreSQLInstance) -> WALSenderState | None:
    """Return the state of the WAL sender process (on the primary) connected
    to standby 'instance'.

    This queries pg_stat_replication view on the primary, filtered by
    application_name assuming that the standby instance name is used there.
    Prior to PostgreSQL version 12, application_name is always 'walreceiver',
    so this does not work. Otherwise, we retrieve application_name if set in
    primary_conninfo, use cluster_name otherwise or fall back to instance
    name.
    """
    assert instance.standby is not None, f"{instance} is not a standby"
    primary_conninfo = conninfo_to_dict(instance.standby.primary_conninfo)
    try:
        application_name = primary_conninfo["application_name"]
    except KeyError:
        application_name = instance.config().get("cluster_name", instance.name)
    try:
        with db.primary_connect(instance.standby) as cnx, cnx.cursor(
            row_factory=args_row(WALSenderState)
        ) as cur:
            return cur.execute(
                "SELECT state FROM pg_stat_replication WHERE application_name = %s",
                (application_name,),
            ).fetchone()
    except psycopg.OperationalError as e:
        logger.warning("failed to connect to primary: %s", e)
        return None


def libpq_environ(
    instance: BaseInstance, role: str, *, base: dict[str, str] | None = None
) -> dict[str, str]:
    """Return a dict with libpq environment variables for authentication."""
    auth = instance._settings.postgresql.auth
    if base is None:
        env = os.environ.copy()
    else:
        env = base.copy()
    if auth.passfile is not None:
        env.setdefault("PGPASSFILE", str(auth.passfile))
    if auth.password_command and "PGPASSWORD" not in env:
        try:
            cmd_args = [
                c.format(instance=instance, role=role) for c in auth.password_command
            ]
        except ValueError as e:
            raise exceptions.SettingsError(
                f"failed to format auth.password_command: {e}"
            ) from None
        password = cmd.run(cmd_args, check=True).stdout.strip()
        if password:
            env["PGPASSWORD"] = password
    return env



from __future__ import annotations

from .build import Patroni
from .interface import ClusterMember, ServiceManifest
from .system import Service

__all__ = [
    "Patroni",
    "ClusterMember",
    "ServiceManifest",
    "Service",
]



from __future__ import annotations

from collections.abc import Sequence
from typing import IO, Any

import attrs
import click
from pydantic.utils import deep_update

from .. import plugin_manager, postgresql, privileges, roles
from ..ctx import Context
from ..models import helpers, interface, system
from ..settings import Settings
from .util import (
    Callback,
    CompositeCommandGroup,
    OutputFormat,
    dry_run_option,
    instance_identifier_option,
    output_format_option,
    pass_ctx,
    pass_instance,
    print_argspec,
    print_json_for,
    print_schema,
    print_table_for,
)


def print_role_schema(
    context: click.Context, param: click.Parameter, value: bool
) -> None:
    p = plugin_manager(context.obj.ctx.settings)
    model = interface.Role.composite(p)
    return print_schema(context, param, value, model=model)


def print_role_argspec(
    context: click.Context, param: click.Parameter, value: bool
) -> None:
    settings: Settings = context.obj.ctx.settings
    model = interface.Role.composite(plugin_manager(settings))
    print_argspec(context, param, value, model=model)


class RoleCommands(CompositeCommandGroup[interface.Role]):
    """Group for 'role' sub-commands handling some of them that require a
    composite interface.Role model built from registered plugins at
    runtime.
    """

    model = interface.Role


@click.group("role", cls=RoleCommands)
@instance_identifier_option
@click.option(
    "--schema",
    is_flag=True,
    callback=print_role_schema,
    expose_value=False,
    is_eager=True,
    help="Print the JSON schema of role model and exit.",
)
@click.option(
    "--ansible-argspec",
    is_flag=True,
    callback=print_role_argspec,
    expose_value=False,
    is_eager=True,
    hidden=True,
    help="Print the Ansible argspec of role model and exit.",
)
def cli(**kwargs: Any) -> None:
    """Manage roles."""


# Help mypy because click.group() looses the type of 'cls' argument.
assert isinstance(cli, RoleCommands)


@cli.command_with_composite_model("create")
def _create(Role: type[interface.Role]) -> Callback:
    @helpers.parameters_from_model(Role, "create")
    @pass_instance
    @pass_ctx
    def command(ctx: Context, instance: system.Instance, role: interface.Role) -> None:
        """Create a role in a PostgreSQL instance"""
        with ctx.lock, postgresql.running(ctx, instance):
            if roles.exists(ctx, instance, role.name):
                raise click.ClickException("role already exists")
            roles.apply(ctx, instance, role)

    return command


@cli.command_with_composite_model("alter")
def _alter(Role: type[interface.Role]) -> Callback:
    @helpers.parameters_from_model(Role, "update", parse_model=False)
    @click.argument("rolname")
    @pass_instance
    @pass_ctx
    def command(
        ctx: Context, instance: system.Instance, rolname: str, **changes: Any
    ) -> None:
        """Alter a role in a PostgreSQL instance"""
        with ctx.lock, postgresql.running(ctx, instance):
            values = roles.get(ctx, instance, rolname).dict()
            values = deep_update(values, changes)
            altered = Role.parse_obj(values)
            roles.apply(ctx, instance, altered)

    return command


@cli.command("apply", hidden=True)
@click.option("-f", "--file", type=click.File("r"), metavar="MANIFEST", required=True)
@output_format_option
@dry_run_option
@pass_instance
@pass_ctx
def apply(
    ctx: Context,
    instance: system.Instance,
    file: IO[str],
    output_format: OutputFormat,
    dry_run: bool,
) -> None:
    """Apply manifest as a role"""
    p = plugin_manager(instance._settings)
    model = interface.Role.composite(p)
    role = model.parse_yaml(file)
    if dry_run:
        ret = interface.ApplyResult(change_state=None)
    else:
        with ctx.lock, postgresql.running(ctx, instance):
            ret = roles.apply(ctx, instance, role)
    if output_format == OutputFormat.json:
        print_json_for(ret)


@cli.command("list")
@output_format_option
@pass_instance
@pass_ctx
def ls(
    ctx: Context,
    instance: system.Instance,
    output_format: OutputFormat,
) -> None:
    """List roles in instance"""

    with postgresql.running(ctx, instance):
        rls = [
            r.dict(by_alias=True, exclude={"pgpass"}) for r in roles.ls(ctx, instance)
        ]

    if output_format == OutputFormat.json:
        print_json_for(rls)
    else:
        print_table_for(rls)


@cli.command("get")
@output_format_option
@click.argument("name")
@pass_instance
@pass_ctx
def get(
    ctx: Context,
    instance: system.Instance,
    name: str,
    output_format: OutputFormat,
) -> None:
    """Get the description of a role"""
    with postgresql.running(ctx, instance):
        m = roles.get(ctx, instance, name).dict(by_alias=True)
    if output_format == OutputFormat.json:
        print_json_for(m)
    else:
        print_table_for([m], box=None)


@cli.command("drop")
@helpers.parameters_from_model(interface.RoleDropped, "create")
@pass_instance
@pass_ctx
def drop(
    ctx: Context, instance: system.Instance, roledropped: interface.RoleDropped
) -> None:
    """Drop a role"""
    with postgresql.running(ctx, instance):
        roles.drop(ctx, instance, roledropped)


@cli.command("privileges")
@click.argument("name")
@click.option(
    "-d", "--database", "databases", multiple=True, help="Database to inspect"
)
@click.option("--default", "defaults", is_flag=True, help="Display default privileges")
@output_format_option
@pass_instance
@pass_ctx
def list_privileges(
    ctx: Context,
    instance: system.Instance,
    name: str,
    databases: Sequence[str],
    defaults: bool,
    output_format: OutputFormat,
) -> None:
    """List privileges of a role."""
    with postgresql.running(ctx, instance):
        roles.get(ctx, instance, name)  # check existence
        try:
            prvlgs = privileges.get(
                ctx, instance, databases=databases, roles=(name,), defaults=defaults
            )
        except ValueError as e:
            raise click.ClickException(str(e)) from None
    values = [attrs.asdict(p) for p in prvlgs]
    if output_format == OutputFormat.json:
        print_json_for(values)
    else:
        print_table_for(values)



from __future__ import annotations

import logging
from pathlib import Path

import pytest

from pglift import util


def test_xdg_config_home(monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.setattr("pathlib.Path.home", lambda: Path("/ho/me"))
        assert util.xdg_config_home() == Path("/ho/me/.config")


def test_xdg_data_home(monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.setenv("XDG_DATA_HOME", "/x/y")
        assert util.xdg_data_home() == Path("/x/y")
    with monkeypatch.context() as m:
        try:
            m.delenv("XDG_DATA_HOME")
        except KeyError:
            pass
        m.setattr("pathlib.Path.home", lambda: Path("/ho/me"))
        assert util.xdg_data_home() == Path("/ho/me/.local/share")


def test_xdg_config(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    configdir = tmp_path / "pglift"
    configdir.mkdir()
    configfile = configdir / "x"
    configfile.touch()
    with monkeypatch.context() as m:
        m.setenv("XDG_CONFIG_HOME", str(tmp_path))
        assert util.xdg_config("x") is not None
    assert util.xdg_config("x") is None


def test_custom_config(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    configfile = tmp_path / "x"
    configfile.touch()
    with monkeypatch.context() as m:
        m.setenv("PGLIFT_CONFIG_PATH", "/not/existing/path")
        with pytest.raises(Exception, match="does not exist"):
            assert util.custom_config("x") is not None

        m.setenv("PGLIFT_CONFIG_PATH", str(tmp_path))
        assert util.custom_config("x") is not None
    assert util.custom_config("x") is None


def test_dist_config() -> None:
    pg_hba = util.dist_config("postgresql", "pg_hba.conf")
    assert pg_hba is not None


def test_total_memory(meminfo: Path) -> None:
    assert util.total_memory(meminfo) == 6166585344.0


def test_total_memory_error(tmp_path: Path) -> None:
    meminfo = tmp_path / "meminfo"
    meminfo.touch()
    with pytest.raises(Exception, match="could not retrieve memory information from"):
        util.total_memory(meminfo)


def test_rmdir(tmp_path: Path, caplog: pytest.LogCaptureFixture) -> None:
    d = tmp_path / "x"
    subd = d / "y"
    subd.mkdir(parents=True)

    with caplog.at_level(logging.WARNING):
        assert not util.rmdir(d)
    assert "Directory not empty" in caplog.messages[0]
    assert d.exists()

    caplog.clear()

    with caplog.at_level(logging.WARNING):
        assert util.rmdir(subd)
        assert util.rmdir(d)
    assert not caplog.messages
    assert not d.exists()



import enum
from collections.abc import Iterator
from datetime import datetime
from typing import Any, ClassVar, Final, Literal, Optional, TypeVar, Union

import psycopg.conninfo
import pydantic
from pgtoolkit import conf as pgconf
from pydantic import (
    Field,
    PostgresDsn,
    SecretStr,
    ValidationError,
    fields,
    root_validator,
    validator,
)
from pydantic.error_wrappers import ErrorWrapper
from pydantic.fields import ModelField
from pydantic.utils import lenient_issubclass

from .. import settings as s
from .. import types, util
from .._compat import Self
from ..pm import PluginManager
from ..postgresql import Standby
from ..settings import _postgresql as pgs
from ..types import (
    AnsibleConfig,
    AutoStrEnum,
    CLIConfig,
    CompositeManifest,
    Manifest,
    Port,
    ServiceManifest,
    Status,
)

default_port: Final = 5432


def validate_ports(model: pydantic.BaseModel) -> None:
    """Walk fields of 'model', checking those with type Port if their value is
    available.
    """

    def _validate(
        model: pydantic.BaseModel, *, loc: tuple[str, ...] = ()
    ) -> Iterator[ErrorWrapper]:
        cls = model.__class__
        for name, field in cls.__fields__.items():
            if (value := getattr(model, name)) is None:
                continue
            ftype = field.outer_type_
            if lenient_issubclass(ftype, pydantic.BaseModel):
                yield from _validate(value, loc=loc + (name,))
            elif lenient_issubclass(ftype, Port):
                assert isinstance(value, Port)
                if not value.available():
                    yield ErrorWrapper(
                        ValueError(f"port {value} already in use"), loc + (name,)
                    )

    if errors := list(_validate(model)):
        raise ValidationError(errors, model.__class__)


def validate_state_is_absent(
    value: Union[bool, str], values: dict[str, Any], field: fields.ModelField
) -> Union[bool, str]:
    """Make sure state is absent.

    >>> r =  Role(name="bob",  drop_owned=True)
    Traceback (most recent call last):
        ...
    pydantic.error_wrappers.ValidationError: 1 validation error for Role
    drop_owned
      drop_owned can not be set if state is not absent (type=value_error)

    >>> r =  Role(name="bob",  reassign_owned="postgres")
    Traceback (most recent call last):
        ...
    pydantic.error_wrappers.ValidationError: 1 validation error for Role
    reassign_owned
      reassign_owned can not be set if state is not absent (type=value_error)

    >>> r =  Database(name="db1", force_drop=True)
    Traceback (most recent call last):
        ...
    pydantic.error_wrappers.ValidationError: 1 validation error for Database
    force_drop
      force_drop can not be set if state is not absent (type=value_error)
    """
    absent = PresenceState.absent
    if value and values["state"] != absent:
        raise ValueError(f"{field.name} can not be set if state is not {absent}")
    return value


class InstanceState(AutoStrEnum):
    """Instance state."""

    stopped = enum.auto()
    """stopped"""

    started = enum.auto()
    """started"""

    absent = enum.auto()
    """absent"""

    restarted = enum.auto()
    """restarted"""

    @classmethod
    def from_pg_status(cls, status: Status) -> Self:
        """Instance state from PostgreSQL status.

        >>> InstanceState.from_pg_status(Status.running)
        <InstanceState.started: 'started'>
        >>> InstanceState.from_pg_status(Status.not_running)
        <InstanceState.stopped: 'stopped'>
        """
        return cls(
            {
                status.running: cls.started,
                status.not_running: cls.stopped,
            }[status]
        )


class PresenceState(AutoStrEnum):
    """Should the object be present or absent?"""

    present = enum.auto()
    absent = enum.auto()


class BaseInstance(Manifest):
    """PostgreSQL instance suitable for lookup"""

    name: str = Field(readOnly=True, description="Instance name.")
    version: Optional[pgs.PostgreSQLVersion] = Field(
        default=None, description="PostgreSQL version.", readOnly=True
    )

    @validator("name")
    def __validate_name_(cls, v: str) -> str:
        """Validate 'name' field.

        >>> Instance(name='without_dash')  # doctest: +ELLIPSIS
        Instance(name='without_dash', ...)
        >>> Instance(name='with-dash')
        Traceback (most recent call last):
            ...
        pydantic.error_wrappers.ValidationError: 1 validation error for Instance
        name
          instance name must not contain dashes (type=value_error)
        >>> Instance(name='with/slash')
        Traceback (most recent call last):
            ...
        pydantic.error_wrappers.ValidationError: 1 validation error for Instance
        name
          instance name must not contain slashes (type=value_error)
        """
        # Avoid dash as this will break systemd instance unit.
        if "-" in v:
            raise ValueError("instance name must not contain dashes")
        # Likewise, slash messes up with file paths.
        if "/" in v:
            raise ValueError("instance name must not contain slashes")
        return v


class BaseRole(CompositeManifest):
    name: str = Field(readOnly=True, description="Role name.")
    state: PresenceState = Field(
        const=True,
        default=PresenceState.absent,
        description="Whether the role be present or absent.",
    )
    password: Optional[SecretStr] = Field(
        const=True, default=None, description="Role password.", exclude=True
    )
    encrypted_password: Optional[SecretStr] = Field(
        const=True,
        default=None,
        description="Role password, already encrypted.",
        exclude=True,
    )

    @classmethod
    def component_models(cls, pm: PluginManager) -> list[tuple[str, Any, Any]]:
        return pm.hook.role_model()  # type: ignore[no-any-return]

    @validator("password", "encrypted_password")
    def __validate_password_(
        cls, value: Optional[SecretStr], values: dict[str, Any], field: ModelField
    ) -> Optional[SecretStr]:
        """Make sure 'password' and 'encrypted_password' are not specified together.

        >>> Role(name="bob", password="secret", encrypted_password="tercec")
        Traceback (most recent call last):
            ...
        pydantic.error_wrappers.ValidationError: 1 validation error for Role
        encrypted_password
          field is mutually exclusive with 'password' (type=value_error)

        >>> r = Role(name="bob", encrypted_password="tercec")
        >>> r.password, r.encrypted_password
        (None, SecretStr('**********'))
        """
        other = (
            "password" if field.name == "encrypted_password" else "encrypted_password"
        )
        if value and values.get(other):
            raise ValueError(f"field is mutually exclusive with {other!r}")
        return value


class _RoleExisting(BaseRole):
    """Base model for a role that exists (or should exist, after creation)."""

    _cli_config: ClassVar[dict[str, CLIConfig]] = {
        "in_roles": {"name": "in_role"},
        "state": {"hide": True},
        "has_password": {"hide": True},
    }
    _ansible_config: ClassVar[dict[str, AnsibleConfig]] = {
        "has_password": {"hide": True},
    }

    password: Optional[SecretStr] = Field(
        default=None, description="Role password.", exclude=True
    )
    encrypted_password: Optional[SecretStr] = Field(
        default=None, description="Role password, already encrypted.", exclude=True
    )
    has_password: bool = Field(
        default=False,
        description="True if the role has a password.",
        readOnly=True,
    )
    inherit: bool = Field(
        default=True,
        description="Let the role inherit the privileges of the roles it is a member of.",
    )
    login: bool = Field(default=False, description="Allow the role to log in.")
    superuser: bool = Field(
        default=False, description="Whether the role is a superuser."
    )
    createdb: bool = Field(
        default=False, description="Whether role can create new databases."
    )
    createrole: bool = Field(
        default=False, description="Whether role can create new roles."
    )
    replication: bool = Field(
        default=False, description="Whether the role is a replication role."
    )
    connection_limit: Optional[int] = Field(
        description="How many concurrent connections the role can make.",
        default=None,
    )
    validity: Optional[datetime] = Field(
        description="Date and time after which the role's password is no longer valid.",
        default=None,
    )
    in_roles: list[str] = Field(
        default=[],
        description="List of roles to which the new role will be added as a new member.",
    )
    state: PresenceState = Field(
        default=PresenceState.present,
        description="Whether the role be present or absent.",
        exclude=True,
    )

    @validator("has_password", always=True)
    def __set_has_password(cls, value: bool, values: dict[str, Any]) -> bool:
        """Set 'has_password' field according to 'password'.

        >>> r = Role(name="postgres")
        >>> r.has_password
        False
        >>> r = Role(name="postgres", password="P4zzw0rd")
        >>> r.has_password
        True
        >>> r = Role(name="postgres", has_password=True)
        >>> r.has_password
        True
        """
        return (
            value
            or values["password"] is not None
            or values["encrypted_password"] is not None
        )


class RoleDropped(BaseRole):
    """Model for a role that is being dropped."""

    drop_owned: bool = Field(
        default=False,
        description="Drop all PostgreSQL's objects owned by the role being dropped.",
        exclude=True,
    )
    reassign_owned: Optional[str] = Field(
        default=None,
        description="Reassign all PostgreSQL's objects owned by the role being dropped to the specified role name.",
        min_length=1,
        exclude=True,
    )

    @validator("reassign_owned")
    def __validate_reassign_owned(cls, value: str, values: dict[str, Any]) -> str:
        """Validate reassign_owned fields.

        >>> r = RoleDropped(name="bob", drop_owned=True, reassign_owned="postgres")
        Traceback (most recent call last):
            ...
        pydantic.error_wrappers.ValidationError: 1 validation error for RoleDropped
        reassign_owned
          drop_owned and reassign_owned are mutually exclusive (type=value_error)

        >>> r = RoleDropped(name="bob", reassign_owned="")
        Traceback (most recent call last):
            ...
        pydantic.error_wrappers.ValidationError: 1 validation error for RoleDropped
        reassign_owned
          ensure this value has at least 1 characters (type=value_error.any_str.min_length; limit_value=1)
        >>> RoleDropped(name="bob", reassign_owned=None, drop_owned=True)  # doctest: +ELLIPSIS
        RoleDropped(name='bob', state=<PresenceState.absent: 'absent'>, ..., drop_owned=True, reassign_owned=None)
        """
        if value and values["drop_owned"]:
            raise ValueError("drop_owned and reassign_owned are mutually exclusive")
        return value


class Role(_RoleExisting, RoleDropped):
    """PostgreSQL role"""

    _cli_config: ClassVar[dict[str, CLIConfig]] = _RoleExisting._cli_config | {
        "drop_owned": {"hide": True},
        "reassign_owned": {"hide": True},
    }

    _validate_drop_owned = validator("drop_owned", allow_reuse=True)(
        validate_state_is_absent
    )
    _validate_reassign_owned = validator("reassign_owned", allow_reuse=True)(
        validate_state_is_absent
    )


class BaseDatabase(Manifest):
    _cli_config: ClassVar[dict[str, CLIConfig]] = {
        "state": {"hide": True},
    }
    name: str = Field(readOnly=True, description="Database name.", examples=["demo"])

    state: PresenceState = Field(
        const=True,
        default=PresenceState.absent,
        description="Database state.",
        examples=["present"],
    )


class Schema(Manifest):
    name: str = Field(readonly=True, description="Schema name.")

    state: PresenceState = Field(
        default=PresenceState.present,
        description="Schema state.",
        examples=["present"],
        exclude=True,
    )

    owner: Optional[str] = Field(
        description="The role name of the user who will own the schema.",
        examples=["postgres"],
        default=None,
    )


class Extension(Manifest):
    class Config:
        frozen = True

    name: str = Field(readOnly=True, description="Extension name.")
    schema_: Optional[str] = Field(
        alias="schema",
        default=None,
        description="Name of the schema in which to install the extension's object.",
    )
    version: Optional[str] = Field(
        default=None, description="Version of the extension to install."
    )

    state: PresenceState = Field(
        default=PresenceState.present,
        description="Extension state.",
        examples=["present"],
        exclude=True,
    )


class Publication(Manifest):
    name: str = Field(
        description="Name of the publication, unique in the database.",
    )
    state: PresenceState = Field(
        default=PresenceState.present,
        description="Presence state.",
        examples=["present"],
        exclude=True,
    )


class ConnectionString(Manifest):
    conninfo: str = Field(
        description="The libpq connection string, without password.",
    )
    password: Optional[SecretStr] = Field(
        default=None,
        description="Optional password to inject into the connection string.",
        readOnly=True,
        exclude=True,
    )

    @classmethod
    def parse(cls, value: str) -> Self:
        conninfo = psycopg.conninfo.conninfo_to_dict(value)
        password = conninfo.pop("password", None)
        return cls(
            conninfo=psycopg.conninfo.make_conninfo(**conninfo), password=password
        )

    @property
    def full_conninfo(self) -> str:
        """The full connection string, including password field."""
        password = None
        if self.password:
            password = self.password.get_secret_value()
        return psycopg.conninfo.make_conninfo(self.conninfo, password=password)

    @validator("conninfo")
    def __validate_conninfo_(cls, value: str) -> str:
        s = psycopg.conninfo.conninfo_to_dict(value)
        if "password" in s:
            raise ValueError("must not contain a password")
        return psycopg.conninfo.make_conninfo(**{k: v for k, v in sorted(s.items())})


class Subscription(Manifest):
    name: str = Field(description="Name of the subscription.")
    connection: ConnectionString = Field(
        description="The libpq connection string defining how to connect to the publisher database.",
        readOnly=True,
    )
    publications: list[str] = Field(
        description="List of publications on the publisher to subscribe to.",
        min_items=1,
    )
    enabled: bool = Field(
        description="Enable or disable the subscription.",
        default=True,
    )
    state: PresenceState = Field(
        default=PresenceState.present,
        description="Presence state.",
        examples=["present"],
        exclude=True,
    )

    @classmethod
    def from_row(cls, **kwargs: Any) -> Self:
        return cls(
            connection=ConnectionString.parse(kwargs.pop("connection")), **kwargs
        )


class CloneOptions(Manifest):
    _cli_config: ClassVar[dict[str, CLIConfig]] = {
        "dsn": {"name": "from", "metavar": "conninfo"},
    }
    dsn: PostgresDsn = Field(
        description="Data source name of the database to restore into this one, specified as a libpq connection URI.",
    )
    schema_only: bool = Field(
        description="Only restore the schema (data definitions).",
        default=False,
    )


class _DatabaseExisting(BaseDatabase):
    """Base model for a database that exists (or should exist, after creation)."""

    _cli_config: ClassVar[dict[str, CLIConfig]] = {
        "settings": {"hide": True},
        "state": {"hide": True},
        "extensions": {"name": "extension"},
        "publications": {"hide": True},
        "subscriptions": {"hide": True},
        "schemas": {"name": "schema"},
    }
    _ansible_config: ClassVar[dict[str, AnsibleConfig]] = {
        "settings": {"spec": {"type": "dict", "required": False}},
    }

    state: PresenceState = Field(
        default=PresenceState.present,
        description="Database state.",
        examples=["present"],
        exclude=True,
    )
    owner: Optional[str] = Field(
        description="The role name of the user who will own the database.",
        examples=["postgres"],
        default=None,
    )
    settings: Optional[dict[str, Optional[pgconf.Value]]] = Field(
        default=None,
        description=(
            "Session defaults for run-time configuration variables for the database. "
            "Upon update, an empty (dict) value would reset all settings."
        ),
        examples=[{"work_mem": "5MB"}],
    )
    schemas: list[Schema] = Field(
        default=[],
        description="List of schemas to create in the database.",
        examples=[{"name": "sales"}, "accounting"],
    )
    extensions: list[Extension] = Field(
        default=[],
        description="List of extensions to create in the database.",
        examples=[{"name": "unaccent", "schema": "ext", "version": "1.0"}, "hstore"],
    )

    publications: list[Publication] = Field(
        default=[],
        description="List of publications to in the database.",
        examples=[{"name": "mypub"}],
    )

    subscriptions: list[Subscription] = Field(
        default=[],
        description="List of subscriptions to in the database.",
        examples=[
            {"name": "mysub", "publications": ["mypub"], "enabled": False},
        ],
    )

    clone: Optional[CloneOptions] = Field(
        description="Options for cloning a database into this one.",
        readOnly=True,
        writeOnly=True,
        examples=[
            "postgresql://app:password@dbserver:5455/appdb",
            {
                "dsn": "postgresql://app:password@dbserver:5455/appdb",
                "schema_only": True,
            },
        ],
        default=None,
        exclude=True,
    )

    tablespace: Optional[str] = Field(
        description="The name of the tablespace that will be associated with the database.",
        default=None,
    )

    @validator("extensions", each_item=True, pre=True)
    def __validate_extensions_(
        cls, value: Union[str, dict[str, Any]]
    ) -> dict[str, Any]:
        """Turn extension values from strings (name) as full objects.

        >>> Database(name="x", extensions=["named", {"name": "obj", "schema": "sch", "version": "1.5"}])  # doctest: +ELLIPSIS
        Database(name='x', ..., extensions=[Extension(name='named', schema_=None, version=None, ...), Extension(name='obj', schema_='sch', version='1.5'...)], ...)
        """
        if isinstance(value, str):
            return {"name": value}
        return value

    @validator("schemas", each_item=True, pre=True)
    def __validate_schemas_(cls, value: Union[str, dict[str, Any]]) -> dict[str, Any]:
        """Turn schema values from strings (name) as full objects.

        >>> Database(name="x", schemas=["named", {"name": "obj"}])  # doctest: +ELLIPSIS
        Database(name='x', ..., schemas=[Schema(name='named', ...), Schema(name='obj', ...)], ...)
        """
        if isinstance(value, str):
            return {"name": value}
        return value

    @validator("tablespace")
    def __validate_tablespace_(cls, value: str) -> Optional[str]:
        """Convert 'DEFAULT' string for tablespace to None

        >>> Database(name="x", tablespace="default")
        Traceback (most recent call last):
            ...
        pydantic.error_wrappers.ValidationError: 1 validation error for Database
        tablespace
          'default' is not a valid value ... (type=value_error)
        >>> Database(name="x", tablespace="DEFAULT")
        Traceback (most recent call last):
            ...
        pydantic.error_wrappers.ValidationError: 1 validation error for Database
        tablespace
          'DEFAULT' is not a valid value ... (type=value_error)
        """
        if value and value.lower() == "default":
            raise ValueError(
                f"{value!r} is not a valid value for 'tablespace'. "
                "Don't provide a value if you want the tablespace to be set to DEFAULT."
            )
        return value


class DatabaseDropped(BaseDatabase):
    """Model for a database that is being dropped."""

    _cli_config: ClassVar[dict[str, CLIConfig]] = {
        "force_drop": {"name": "force"},
    }

    force_drop: bool = Field(default=False, description="Force the drop.", exclude=True)


class Database(_DatabaseExisting, DatabaseDropped):
    """PostgreSQL database"""

    _cli_config: ClassVar[dict[str, CLIConfig]] = _DatabaseExisting._cli_config | {
        "force_drop": {"hide": True},
    }

    _validate_force_drop = validator("force_drop", allow_reuse=True)(
        validate_state_is_absent
    )


class Auth(types.BaseModel):
    local: Optional[pgs.AuthLocalMethod] = Field(
        default=None,
        description="Authentication method for local-socket connections",
        readOnly=True,
    )
    host: Optional[pgs.AuthHostMethod] = Field(
        default=None,
        description="Authentication method for local TCP/IP connections",
        readOnly=True,
    )
    hostssl: Optional[pgs.AuthHostSSLMethod] = Field(
        default=None,
        description="Authentication method for SSL-encrypted TCP/IP connections",
        readOnly=True,
    )


class Instance(CompositeManifest, BaseInstance):
    """PostgreSQL instance"""

    _cli_config: ClassVar[dict[str, CLIConfig]] = {
        "status": {"hide": True},
        "state": {
            "choices": [InstanceState.started.value, InstanceState.stopped.value]
        },
        "pending_restart": {"hide": True},
        "restart_on_changes": {"hide": True},
        "settings": {"hide": True},
        "roles": {"hide": True},
        "databases": {"hide": True},
    }
    _ansible_config: ClassVar[dict[str, AnsibleConfig]] = {
        "pending_restart": {"hide": True},
    }

    @classmethod
    def component_models(cls, pm: PluginManager) -> list[tuple[str, Any, Any]]:
        return pm.hook.interface_model()  # type: ignore[no-any-return]

    port: Port = Field(
        default=Port(default_port),
        description=(
            "TCP port the postgresql instance will be listening to. "
            f"If unspecified, default to {default_port} unless a 'port' setting is found in 'settings'."
        ),
    )
    settings: dict[str, Any] = Field(
        default={},
        description=("Settings for the PostgreSQL instance."),
        examples=[
            {
                "listen_addresses": "*",
                "shared_buffers": "1GB",
                "ssl": True,
                "ssl_key_file": "/etc/certs/db.key",
                "ssl_cert_file": "/etc/certs/db.key",
                "shared_preload_libraries": "pg_stat_statements",
            }
        ],
    )
    surole_password: Optional[SecretStr] = Field(
        default=None,
        description="Super-user role password.",
        readOnly=True,
        exclude=True,
    )
    replrole_password: Optional[SecretStr] = Field(
        default=None,
        description="Replication role password.",
        readOnly=True,
        exclude=True,
    )
    data_checksums: Optional[bool] = Field(
        default=None,
        description=(
            "Enable or disable data checksums. "
            "If unspecified, fall back to site settings choice."
        ),
    )
    locale: Optional[str] = Field(
        default=None, description="Default locale.", readOnly=True
    )
    encoding: Optional[str] = Field(
        default=None,
        description="Character encoding of the PostgreSQL instance.",
        readOnly=True,
    )

    auth: Optional[Auth] = Field(default=None, exclude=True, writeOnly=True)

    standby: Optional[Standby] = Field(default=None, description="Standby information.")

    state: InstanceState = Field(
        default=InstanceState.started,
        description="Runtime state.",
    )
    databases: list[Database] = Field(
        default=[],
        description="Databases defined in this instance (non-exhaustive list).",
        exclude=True,
        writeOnly=True,
    )
    roles: list[Role] = Field(
        default=[],
        description="Roles defined in this instance (non-exhaustive list).",
        exclude=True,
        writeOnly=True,
    )

    pending_restart: bool = Field(
        default=False,
        description="Whether the instance needs a restart to account for settings changes.",
        readOnly=True,
    )
    restart_on_changes: bool = Field(
        default=False,
        description="Whether or not to automatically restart the instance to account for settings changes.",
        exclude=True,
        writeOnly=True,
    )

    @root_validator(pre=True)
    def __validate_port_(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate that 'port' field and settings['port'] are consistent.

        If unspecified, 'port' is either set from settings value or from
        the default port value.

        >>> i = Instance(name="i")
        >>> i.port, "port" in i.settings
        (5432, False)
        >>> i = Instance(name="i", settings={"port": 5423})
        >>> i.port, i.settings["port"]
        (5423, 5423)
        >>> i = Instance(name="i", port=5454)
        >>> i.port, "port" in i.settings
        (5454, False)

        Otherwise, and if settings['port'] exists, make sure values are
        consistent and possibly cast the latter as an integer.

        >>> i = Instance(name="i", settings={"port": 5455})
        >>> i.port, i.settings["port"]
        (5455, 5455)
        >>> i = Instance(name="i", port=123, settings={"port": "123"})
        >>> i.port, i.settings["port"]
        (123, 123)
        >>> Instance(name="i", port=321, settings={"port": 123})
        Traceback (most recent call last):
            ...
        pydantic.error_wrappers.ValidationError: 1 validation error for Instance
        __root__
          'port' field and settings['port'] mismatch (type=value_error)
        >>> Instance(name="i", settings={"port": "abc"})
        Traceback (most recent call last):
            ...
        pydantic.error_wrappers.ValidationError: 1 validation error for Instance
        __root__
          invalid literal for int() with base 10: 'abc' (type=value_error)
        """
        config_port = None
        try:
            port = values["port"]
        except KeyError:
            try:
                config_port = int(values["settings"]["port"])
            except KeyError:
                pass
            else:
                values["port"] = Port(config_port)
        else:
            try:
                config_port = int(values["settings"]["port"])
            except KeyError:
                pass
            else:
                if config_port != port:
                    raise ValueError("'port' field and settings['port'] mismatch")
        if config_port is not None:
            values["settings"]["port"] = config_port
        return values

    @root_validator
    def __validate_standby_and_patroni_(cls, values: dict[str, Any]) -> dict[str, Any]:
        if values.get("standby") and values.get("patroni"):
            raise ValueError("'patroni' and 'standby' fields are mutually exclusive")
        return values

    _S = TypeVar("_S", bound=ServiceManifest)

    def service_manifest(self, stype: type[_S]) -> _S:
        """Return satellite service manifest attached to this instance.

        :raises ValueError: if not found.
        """
        fname = stype.__service__
        try:
            s = getattr(self, fname)
        except AttributeError as e:
            raise ValueError(fname) from e
        if s is None:
            raise ValueError(fname)
        assert isinstance(
            s, stype
        ), f"expecting field {fname} to have type {stype} (got {type(s)})"
        return s

    def surole(self, settings: s.Settings) -> Role:
        s = settings.postgresql.surole
        extra = {}
        if settings.postgresql.auth.passfile is not None:
            extra["pgpass"] = s.pgpass
        return Role(name=s.name, password=self.surole_password, **extra)

    def replrole(self, settings: s.Settings) -> Optional[Role]:
        if (name := settings.postgresql.replrole) is None:
            return None
        return Role(
            name=name,
            password=self.replrole_password,
            login=True,
            replication=True,
            in_roles=["pg_read_all_stats"],
        )

    def auth_options(self, settings: pgs.AuthSettings) -> Auth:
        local, host, hostssl = settings.local, settings.host, settings.hostssl
        if auth := self.auth:
            local = auth.local or local
            host = auth.host or host
            hostssl = auth.hostssl or hostssl
        return Auth(local=local, host=host, hostssl=hostssl)

    def pg_hba(self, settings: s.Settings) -> str:
        surole = self.surole(settings)
        replrole = self.replrole(settings)
        replrole_name = replrole.name if replrole else None
        auth = self.auth_options(settings.postgresql.auth)
        return util.template("postgresql", "pg_hba.conf").format(
            auth=auth,
            surole=surole.name,
            backuprole=settings.postgresql.backuprole.name,
            replrole=replrole_name,
        )

    def pg_ident(self, settings: s.Settings) -> str:
        surole = self.surole(settings)
        replrole = self.replrole(settings)
        replrole_name = replrole.name if replrole else None
        return util.template("postgresql", "pg_ident.conf").format(
            surole=surole.name,
            backuprole=settings.postgresql.backuprole.name,
            replrole=replrole_name,
            sysuser=settings.sysuser[0],
        )

    def initdb_options(self, base: pgs.InitdbSettings) -> pgs.InitdbSettings:
        data_checksums: Union[None, Literal[True]] = {
            True: True,
            False: None,
            None: base.data_checksums or None,
        }[self.data_checksums]
        return pgs.InitdbSettings(
            locale=self.locale or base.locale,
            encoding=self.encoding or base.encoding,
            data_checksums=data_checksums,
        )


class ApplyChangeState(AutoStrEnum):
    """A apply change state for object handled by pglift"""

    created = enum.auto()  #:
    changed = enum.auto()  #:
    dropped = enum.auto()  #:


class ApplyResult(Manifest):
    """
    ApplyResult allows to describe the result of a call to apply function
    (Eg: pglift.database.apply) to an object (Eg: database, instance,...).

    The `change_state` attribute of this class can be set to one of to those values:
      - :attr:`~ApplyChangeState.created` if the object has been created,
      - :attr:`~ApplyChangeState.changed` if the object has been changed,
      - :attr:`~ApplyChangeState.dropped` if the object has been dropped,
      - :obj:`None` if nothing happened to the object we manipulate (neither created,
        changed or dropped)
    """

    change_state: Optional[ApplyChangeState] = Field(
        description="Define the change applied (created, changed or dropped) to a manipulated object",
    )  #:


class InstanceApplyResult(ApplyResult):
    pending_restart: bool = Field(
        default=False,
        description="Whether the instance needs a restart to account for settings changes.",
    )



from __future__ import annotations

import datetime
import logging
import shlex
import subprocess
from collections.abc import Sequence
from typing import Any

import psycopg.rows
import pydantic
from pgtoolkit import conf as pgconf
from psycopg import sql

from . import (
    cmd,
    db,
    exceptions,
    extensions,
    hookimpl,
    publications,
    schemas,
    subscriptions,
    types,
)
from .ctx import Context
from .models import interface, system
from .postgresql.ctl import libpq_environ
from .task import task

logger = logging.getLogger(__name__)


def apply(
    ctx: Context,
    instance: system.PostgreSQLInstance,
    database: interface.Database,
) -> interface.ApplyResult:
    """Apply state described by specified interface model as a PostgreSQL database.

    The instance should be running and not a standby.
    """
    if instance.standby:
        raise exceptions.InstanceReadOnlyError(instance)

    with db.connect(instance, ctx=ctx) as cnx:
        return _apply(cnx, database, instance, ctx)


def _apply(
    cnx: db.Connection,
    database: interface.Database,
    instance: system.PostgreSQLInstance,
    ctx: Context,
) -> interface.ApplyResult:
    name = database.name
    if database.state == interface.PresenceState.absent:
        dropped = False
        if _exists(cnx, name):
            _drop(cnx, database)
            dropped = True
        return interface.ApplyResult(
            change_state=interface.ApplyChangeState.dropped if dropped else None
        )

    changed = created = False
    if not _exists(cnx, name):
        create(cnx, database, instance)
        created = True
    else:
        logger.info("altering '%s' database on instance %s", database.name, instance)
        changed = alter(cnx, database)

    if (
        database.schemas
        or database.extensions
        or database.publications
        or database.subscriptions
    ):
        with db.connect(instance, ctx=ctx, dbname=name) as db_cnx:
            for schema in database.schemas:
                if schemas.apply(db_cnx, schema, name):
                    changed = True
            for extension in database.extensions:
                if extensions.apply(db_cnx, extension):
                    changed = True
            for publication in database.publications:
                if publications.apply(db_cnx, publication, name):
                    changed = True
            for subscription in database.subscriptions:
                if subscriptions.apply(db_cnx, subscription, name):
                    changed = True

    if created:
        state = interface.ApplyChangeState.created
    elif changed:
        state = interface.ApplyChangeState.changed
    else:
        state = None
    return interface.ApplyResult(change_state=state)


def clone(
    name: str,
    options: interface.CloneOptions,
    instance: system.PostgreSQLInstance,
) -> None:
    logger.info("cloning '%s' database in %s from %s", name, instance, options.dsn)

    def log_cmd(cmd_args: list[str]) -> None:
        args = [
            db.obfuscate_conninfo(a)
            if isinstance(a, (types.ConnectionString, pydantic.PostgresDsn))
            else a
            for a in cmd_args
        ]
        logger.debug(shlex.join(args))

    pg_dump = instance.bindir / "pg_dump"
    dump_cmd = [str(pg_dump), "--format", "custom", "-d", options.dsn]
    user = instance._settings.postgresql.surole.name
    restore_cmd = [
        str(instance.bindir / "pg_restore"),
        "--exit-on-error",
        "-d",
        db.dsn(instance, dbname=name, user=user),
    ]
    if options.schema_only:
        dump_cmd.append("--schema-only")
        restore_cmd.append("--schema-only")
    env = libpq_environ(instance, user)
    with subprocess.Popen(  # nosec
        dump_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
    ) as dump:
        log_cmd(dump_cmd)
        restore = cmd.run(restore_cmd, stdin=dump.stdout, env=env)
        pg_dump_stderr = []
        assert dump.stderr
        for errline in dump.stderr:
            logger.debug("%s: %s", pg_dump, errline.rstrip())
            pg_dump_stderr.append(errline)

    if dump.returncode:
        raise exceptions.CommandError(
            dump.returncode, dump_cmd, stderr="".join(pg_dump_stderr)
        )
    if restore.returncode:
        raise exceptions.CommandError(
            restore.returncode, restore_cmd, restore.stdout, restore.stderr
        )


def get(
    ctx: Context, instance: system.PostgreSQLInstance, name: str
) -> interface.Database:
    """Return the database object with specified name.

    :raises ~pglift.exceptions.DatabaseNotFound: if no database with specified
        'name' exists.
    """
    if not exists(ctx, instance, name):
        raise exceptions.DatabaseNotFound(name)
    with db.connect(instance, ctx=ctx, dbname=name) as cnx:
        return _get(cnx, dbname=name)


def _get(cnx: db.Connection, dbname: str) -> interface.Database:
    row = cnx.execute(db.query("database_inspect"), {"database": dbname}).fetchone()
    assert row is not None
    settings = row.pop("settings")
    if settings is None:
        row["settings"] = None
    else:
        row["settings"] = {}
        for s in settings:
            k, v = s.split("=", 1)
            row["settings"][k.strip()] = pgconf.parse_value(v.strip())
    row["schemas"] = schemas.ls(cnx)
    row["extensions"] = extensions.ls(cnx)
    row["publications"] = publications.ls(cnx)
    row["subscriptions"] = subscriptions.ls(cnx, dbname)
    return interface.Database.parse_obj(row)


def ls(
    ctx: Context, instance: system.PostgreSQLInstance, dbnames: Sequence[str] = ()
) -> list[system.Database]:
    """List databases in instance.

    :param dbnames: restrict operation on databases with a name in this list.
    """
    with db.connect(instance, ctx=ctx) as cnx:
        return _list(cnx, dbnames)


def _list(cnx: db.Connection, dbnames: Sequence[str] = ()) -> list[system.Database]:
    where_clause = (
        sql.SQL("AND d.datname IN ({})").format(
            sql.SQL(", ").join(map(sql.Literal, dbnames))
        )
        if dbnames
        else sql.SQL("")
    )
    with cnx.cursor(row_factory=psycopg.rows.kwargs_row(system.Database.build)) as cur:
        cur.execute(db.query("database_list", where_clause=where_clause))
        return cur.fetchall()


def drop(
    ctx: Context,
    instance: system.PostgreSQLInstance,
    database: interface.DatabaseDropped,
) -> None:
    """Drop a database from a primary instance.

    :raises ~pglift.exceptions.DatabaseNotFound: if no database with specified
        'name' exists.
    """
    if instance.standby:
        raise exceptions.InstanceReadOnlyError(instance)
    with db.connect(instance, ctx=ctx) as cnx:
        if not _exists(cnx, database.name):
            raise exceptions.DatabaseNotFound(database.name)
        _drop(cnx, database)


def _drop(cnx: db.Connection, database: interface.DatabaseDropped) -> None:
    logger.info("dropping '%s' database", database.name)
    options = ""
    if database.force_drop:
        if cnx.info.server_version < 130000:
            raise exceptions.UnsupportedError(
                "Force drop option can't be used with PostgreSQL < 13"
            )
        options = "WITH (FORCE)"

    cnx.execute(
        db.query(
            "database_drop",
            database=sql.Identifier(database.name),
            options=sql.SQL(options),
        )
    )


def exists(ctx: Context, instance: system.PostgreSQLInstance, name: str) -> bool:
    """Return True if named database exists in 'instance'.

    The instance should be running.
    """
    with db.connect(instance, ctx=ctx) as cnx:
        return _exists(cnx, name)


def _exists(cnx: db.Connection, name: str) -> bool:
    cur = cnx.execute(db.query("database_exists"), {"database": name})
    return cur.rowcount == 1


@task(title="creating '{database.name}' database in {instance}")
def create(
    cnx: db.Connection,
    database: interface.Database,
    instance: system.PostgreSQLInstance,
) -> None:
    opts = []
    if database.owner is not None:
        opts.append(sql.SQL("OWNER {}").format(sql.Identifier(database.owner)))
    if database.tablespace is not None:
        opts.append(
            sql.SQL("TABLESPACE {}").format(sql.Identifier(database.tablespace))
        )

    cnx.execute(
        db.query(
            "database_create",
            database=sql.Identifier(database.name),
            options=sql.SQL(" ").join(opts),
        ),
    )
    if database.settings is not None:
        _configure(cnx, database.name, database.settings)

    if database.clone:
        clone(database.name, database.clone, instance)


@create.revert
def revert_create(
    cnx: db.Connection,
    database: interface.Database,
    instance: system.PostgreSQLInstance,
) -> None:
    _drop(cnx, interface.DatabaseDropped(name=database.name))


def alter(cnx: db.Connection, database: interface.Database) -> bool:
    owner: sql.Composable
    actual = _get(cnx, database.name)
    if database.owner is None:
        owner = sql.SQL("CURRENT_USER")
    else:
        owner = sql.Identifier(database.owner)
    options = sql.SQL("OWNER TO {}").format(owner)
    cnx.execute(
        db.query(
            "database_alter",
            database=sql.Identifier(database.name),
            options=options,
        ),
    )

    if database.settings is not None:
        _configure(cnx, database.name, database.settings)

    if actual.tablespace != database.tablespace and database.tablespace is not None:
        options = sql.SQL("SET TABLESPACE {}").format(
            sql.Identifier(database.tablespace)
        )
        cnx.execute(
            db.query(
                "database_alter",
                database=sql.Identifier(database.name),
                options=options,
            ),
        )

    return _get(cnx, database.name) != actual


def _configure(
    cnx: db.Connection, dbname: str, db_settings: dict[str, pgconf.Value | None]
) -> None:
    if not db_settings:
        # Empty input means reset all.
        cnx.execute(
            db.query(
                "database_alter",
                database=sql.Identifier(dbname),
                options=sql.SQL("RESET ALL"),
            )
        )
    else:
        with cnx.transaction():
            for k, v in db_settings.items():
                if v is None:
                    options = sql.SQL("RESET {}").format(sql.Identifier(k))
                else:
                    options = sql.SQL("SET {} TO {}").format(
                        sql.Identifier(k), sql.Literal(v)
                    )
                cnx.execute(
                    db.query(
                        "database_alter",
                        database=sql.Identifier(dbname),
                        options=options,
                    )
                )


def encoding(cnx: db.Connection) -> str:
    """Return the encoding of connected database."""
    row = cnx.execute(db.query("database_encoding")).fetchone()
    assert row is not None
    value = row["encoding"]
    return str(value)


def run(
    ctx: Context,
    instance: system.PostgreSQLInstance,
    sql_command: str,
    *,
    dbnames: Sequence[str] = (),
    exclude_dbnames: Sequence[str] = (),
    notice_handler: types.NoticeHandler = db.default_notice_handler,
) -> dict[str, list[dict[str, Any]]]:
    """Execute a SQL command on databases of `instance`.

    :param dbnames: restrict operation on databases with a name in this list.
    :param exclude_dbnames: exclude databases with a name in this list from
        the operation.
    :param notice_handler: a function to handle notice.

    :returns: a dict mapping database names to query results, if any.

    :raises psycopg.ProgrammingError: in case of unprocessable query.
    """
    result = {}
    if dbnames:
        target = ", ".join(dbnames)
    else:
        target = "ALL databases"
        if exclude_dbnames:
            target += f" except {', '.join(exclude_dbnames)}"
    if not ctx.confirm(
        f"Confirm execution of {sql_command!r} on {target} of {instance}?", True
    ):
        raise exceptions.Cancelled(f"execution of {sql_command!r} cancelled")

    for database in ls(ctx, instance):
        if (
            dbnames and database.name not in dbnames
        ) or database.name in exclude_dbnames:
            continue
        with db.connect(instance, ctx=ctx, dbname=database.name) as cnx:
            cnx.add_notice_handler(notice_handler)
            logger.info(
                'running "%s" on %s database of %s',
                sql_command,
                database.name,
                instance,
            )
            cur = cnx.execute(sql_command)
            if cur.statusmessage:
                logger.info(cur.statusmessage)
            if cur.description is not None:
                result[database.name] = cur.fetchall()
    return result


def dump(ctx: Context, instance: system.PostgreSQLInstance, dbname: str) -> None:
    """dump a database of `instance` (logical backup)."""
    logger.info("backing up database '%s' on instance %s", dbname, instance)
    if not exists(ctx, instance, dbname):
        raise exceptions.DatabaseNotFound(dbname)
    postgresql_settings = instance._settings.postgresql

    conninfo = db.dsn(instance, dbname=dbname, user=postgresql_settings.surole.name)

    date = (
        datetime.datetime.now(datetime.timezone.utc)
        .astimezone()
        .isoformat(timespec="seconds")
    )
    dumps_directory = instance.dumps_directory
    cmds = [
        [
            c.format(
                bindir=instance.bindir,
                path=dumps_directory,
                conninfo=conninfo,
                dbname=dbname,
                date=date,
            )
            for c in cmd_args
        ]
        for cmd_args in postgresql_settings.dump_commands
    ]
    env = libpq_environ(instance, postgresql_settings.surole.name)
    for cmd_args in cmds:
        cmd.run(cmd_args, check=True, env=env)


@hookimpl
def instance_configured(
    ctx: Context, manifest: interface.Instance, creating: bool
) -> None:
    if creating:
        instance = system.BaseInstance.get(
            manifest.name, manifest.version, ctx.settings
        )
        instance.dumps_directory.mkdir(parents=True, exist_ok=True)


@hookimpl
def instance_dropped(ctx: Context, instance: system.Instance) -> None:
    dumps_directory = instance.dumps_directory
    if not dumps_directory.exists():
        return
    has_dumps = next(dumps_directory.iterdir(), None) is not None
    if not has_dumps or ctx.confirm(
        f"Confirm deletion of database dump(s) for instance {instance}?",
        True,
    ):
        ctx.rmtree(dumps_directory)



from __future__ import annotations

import pytest

from pglift import instances, systemd
from pglift.ctx import Context
from pglift.models import system
from pglift.systemd import scheduler


@pytest.mark.usefixtures("require_systemd_scheduler", "require_pgbackrest_localrepo")
def test_systemd_backup_job(ctx: Context, instance: system.Instance) -> None:
    unit = scheduler.unit("backup", instance.qualname)
    assert systemd.is_enabled(ctx, unit)
    assert systemd.is_active(ctx, unit)
    with instances.stopped(ctx, instance):
        assert not systemd.is_active(ctx, unit)
    assert systemd.is_active(ctx, unit)



from __future__ import annotations

import copy
import json
from pathlib import Path

import pydantic
import pytest
import yaml
from ansible.module_utils.common.arg_spec import ArgumentSpecValidator

from pglift import prometheus
from pglift.models import helpers, interface
from pglift.pm import PluginManager
from pglift.types import Manifest

all_plugins = PluginManager.all()


def test_argspec_from_instance_manifest(
    datadir: Path,
    write_changes: bool,
    composite_instance_model: type[interface.Instance],
) -> None:
    compare_argspec(composite_instance_model, write_changes, datadir)


def test_argspec_from_role_manifest(
    datadir: Path,
    write_changes: bool,
    composite_role_model: type[interface.Role],
) -> None:
    compare_argspec(composite_role_model, write_changes, datadir)


@pytest.mark.parametrize(
    "manifest_type",
    [
        prometheus.PostgresExporter,
        interface.Database,
    ],
)
def test_argspec_from_model_manifest(
    datadir: Path, write_changes: bool, manifest_type: type[Manifest]
) -> None:
    compare_argspec(manifest_type, write_changes, datadir)


def compare_argspec(
    manifest_type: type[Manifest],
    write_changes: bool,
    datadir: Path,
    *,
    name: str | None = None,
) -> None:
    actual = helpers.argspec_from_model(manifest_type)
    if name is None:
        name = manifest_type.__name__.lower()
    fpath = datadir / f"ansible-argspec-{name}.json"
    if write_changes:
        fpath.write_text(json.dumps(actual, indent=2, sort_keys=True) + "\n")
    expected = json.loads(fpath.read_text())
    assert actual == expected


@pytest.mark.parametrize(
    "objtype",
    [
        ("instance", interface.Instance, interface.InstanceApplyResult),
        ("role", interface.Role, interface.ApplyResult),
        ("database", interface.Database, interface.ApplyResult),
        ("postgresexporter", prometheus.models.PostgresExporter, interface.ApplyResult),
    ],
    ids=lambda v: v[0],
)
def test_doc_fragments(
    datadir: Path,
    objtype: tuple[str, type[pydantic.BaseModel], type[pydantic.BaseModel]],
    write_changes: bool,
) -> None:
    name, m, r = objtype
    if hasattr(m, "composite"):
        model = m.composite(all_plugins)
    else:
        model = m
    options = helpers.argspec_from_model(model)
    validator = ArgumentSpecValidator(copy.deepcopy(options))
    examples = (datadir / "ansible-examples" / f"{name}.yaml").read_text()
    for example in yaml.safe_load(examples):
        assert len(example) == 2 and "name" in example
        collection_name = (set(example) - {"name"}).pop()
        assert not validator.validate(example[collection_name]).error_messages
    data = {
        "options": options,
        "return values": helpers.argspec_from_model(r),
        "examples": examples,
    }
    fpath = datadir / "ansible-doc-fragments" / f"{name}.json"
    fpath.parent.mkdir(parents=True, exist_ok=True)
    if write_changes:
        fpath.write_text(json.dumps(data, indent=2, sort_keys=True) + "\n")
    expected = json.loads(fpath.read_text())
    assert data == expected



from __future__ import annotations

import configparser
import json
import logging
from pathlib import Path

import pydantic
from pgtoolkit import conf as pgconf

from .. import exceptions
from .. import service as service_mod
from ..ctx import Context
from ..models import interface, system
from ..settings import Settings, _temboard
from ..task import task
from .models import Service, ServiceManifest, service_name

logger = logging.getLogger(__name__)


def available(settings: Settings) -> _temboard.Settings | None:
    return settings.temboard


def get_settings(settings: Settings) -> _temboard.Settings:
    """Return settings for temboard

    Same as `available` but assert that settings are not None.
    Should be used in a context where settings for the plugin are surely
    set (for example in hookimpl).
    """
    assert settings.temboard is not None
    return settings.temboard


def enabled(qualname: str, settings: _temboard.Settings) -> bool:
    return _configpath(qualname, settings).exists()


def _args(execpath: Path, configpath: Path) -> list[str]:
    return [str(execpath), "--config", str(configpath)]


def _configpath(qualname: str, settings: _temboard.Settings) -> Path:
    return Path(str(settings.configpath).format(name=qualname))


def _homedir(qualname: str, settings: _temboard.Settings) -> Path:
    return Path(str(settings.home).format(name=qualname))


def _pidfile(qualname: str, settings: _temboard.Settings) -> Path:
    return Path(str(settings.pid_file).format(name=qualname))


def _logfile(qualname: str, settings: _temboard.Settings) -> Path:
    return settings.logpath / f"temboard_agent_{qualname}.log"


def config_var(configpath: Path, *, name: str, section: str) -> str:
    """Return temboardagent configuration value for given 'name' in 'section'."""
    if not configpath.exists():
        raise exceptions.FileNotFoundError(
            f"temboard agent configuration file {configpath} not found"
        )
    cp = configparser.ConfigParser()
    cp.read(configpath)
    for s, items in cp.items():
        if s != section:
            continue
        try:
            return items[name]
        except KeyError:
            pass
    raise exceptions.ConfigurationError(
        configpath, f"{name} not found in {section} section"
    )


def port(qualname: str, settings: _temboard.Settings) -> int:
    configpath = _configpath(qualname, settings)
    return int(config_var(configpath, name="port", section="temboard"))


def password(qualname: str, settings: _temboard.Settings) -> str | None:
    configpath = _configpath(qualname, settings)
    try:
        return config_var(configpath, name="password", section="postgresql")
    except exceptions.ConfigurationError:
        return None


def system_lookup(name: str, settings: _temboard.Settings) -> Service | None:
    try:
        p_ = port(name, settings)
        passwd_ = password(name, settings)
    except exceptions.FileNotFoundError as exc:
        logger.debug("failed to read temboard-agent port %s: %s", name, exc)
        return None
    else:
        password_ = None
        if passwd_ is not None:
            password_ = pydantic.SecretStr(passwd_)
        return Service(name=name, settings=settings, port=p_, password=password_)


@task
def setup(
    ctx: Context,
    manifest: interface.Instance,
    settings: _temboard.Settings,
    instance_config: pgconf.Configuration,
) -> None:
    """Setup temboardAgent"""
    service_manifest = manifest.service_manifest(ServiceManifest)

    instance = system.PostgreSQLInstance.system_lookup(
        (manifest.name, manifest.version, ctx.settings)
    )

    configpath = _configpath(instance.qualname, settings)

    password_: str | None = None
    if not configpath.exists():
        if service_manifest.password:
            password_ = service_manifest.password.get_secret_value()
    else:
        # Get the password from config file
        password_ = password(instance.qualname, settings)

    cp = configparser.ConfigParser()
    cp["temboard"] = {
        "port": str(service_manifest.port),
        "plugins": json.dumps(settings.plugins),
        "ssl_ca_cert_file": str(settings.certificate.ca_cert),
        "ssl_cert_file": str(settings.certificate.cert),
        "ssl_key_file": str(settings.certificate.key),
        "home": str(_homedir(instance.qualname, settings)),
        "ui_url": settings.ui_url,
        "signing_public_key": str(settings.signing_key),
    }

    cp["postgresql"] = {
        "user": settings.role,
        "instance": instance.qualname,
    }
    if "port" in instance_config:
        cp["postgresql"]["port"] = str(instance_config["port"])
    if "unix_socket_directories" in instance_config:
        pghost = instance_config.unix_socket_directories.split(",")[0]  # type: ignore[union-attr]
        cp["postgresql"]["host"] = pghost
    if password_:
        cp["postgresql"]["password"] = password_
    cp["logging"] = {"method": settings.logmethod, "level": settings.loglevel}
    if settings.logmethod == _temboard.LogMethod.file:
        cp["logging"]["destination"] = str(_logfile(instance.qualname, settings))

    name = instance.qualname
    needs_restart = False
    if service := system_lookup(name, settings):
        assert configpath.exists()
        cp_actual = configparser.ConfigParser()
        cp_actual.read(configpath)
        if cp != cp_actual:
            logger.info("reconfiguring temboard agent %s", name)
            with configpath.open("w") as configfile:
                cp.write(configfile)
            needs_restart = True
    else:
        logger.info("configuring temboard agent %s", name)
        configpath.parent.mkdir(mode=0o700, exist_ok=True, parents=True)
        configpath.touch(mode=0o600)
        with configpath.open("w") as configfile:
            cp.write(configfile)

        homedir = _homedir(name, settings)
        homedir.mkdir(mode=0o700, exist_ok=True, parents=True)

        pidfile = _pidfile(name, settings)
        pidfile.parent.mkdir(mode=0o700, exist_ok=True, parents=True)

        service = system_lookup(name, settings)
        assert service is not None

    ctx.hook.enable_service(ctx=ctx, service=service_name, name=instance.qualname)

    if needs_restart:
        restart(ctx, service)


@setup.revert(title="deconfiguring temboard agent")
def revert_setup(
    ctx: Context,
    manifest: interface.Instance,
    settings: _temboard.Settings,
    instance_config: pgconf.Configuration,
) -> None:
    """Un-setup temboard"""
    instance = system.BaseInstance.get(manifest.name, manifest.version, ctx.settings)
    _configpath(instance.qualname, settings).unlink(missing_ok=True)
    _pidfile(instance.qualname, settings).unlink(missing_ok=True)
    _logfile(instance.qualname, settings).unlink(missing_ok=True)
    homedir = _homedir(instance.qualname, settings)
    if homedir.exists():
        ctx.rmtree(homedir)
    ctx.hook.disable_service(
        ctx=ctx, service=service_name, name=instance.qualname, now=True
    )


def start(ctx: Context, service: Service, *, foreground: bool = False) -> None:
    logger.info("starting temboard-agent %s", service.name)
    service_mod.start(ctx, service, foreground=foreground)


def stop(ctx: Context, service: Service) -> None:
    logger.info("stopping temboard-agent %s", service.name)
    service_mod.stop(ctx, service)


def restart(ctx: Context, service: Service) -> None:
    logger.info("restarting temboard-agent %s", service.name)
    service_mod.restart(ctx, service)



from __future__ import annotations

from pathlib import Path
from unittest import mock

import pytest

from pglift import systemd
from pglift.settings import Settings, _postgresql, _systemd, default_sysuser
from pglift.types import CompletedProcess


@pytest.mark.usefixtures("systemctl")
def test_executeas(postgresql_settings: _postgresql.Settings) -> None:
    s = Settings.parse_obj(
        {"systemd": {"user": True}, "postgresql": postgresql_settings.dict()}
    )
    assert systemd.executeas(s) == ""

    s = Settings.parse_obj(
        {
            "systemd": {"user": False},
            "sysuser": ["postgres", "pgsql"],
            "postgresql": postgresql_settings,
        }
    )
    assert systemd.executeas(s) == "User=postgres\nGroup=pgsql"


def test_systemctl_cmd(systemctl: str) -> None:
    settings = _systemd.Settings()
    assert systemd.systemctl_cmd(settings, "status", "-n", "2", unit="foo.service") == [
        systemctl,
        "--user",
        "-n",
        "2",
        "status",
        "foo.service",
    ]

    settings = _systemd.Settings(user=False)
    assert systemd.systemctl_cmd(settings, "daemon-reload", unit=None) == [
        systemctl,
        "--system",
        "daemon-reload",
    ]

    settings = _systemd.Settings(user=False, sudo=True)
    assert systemd.systemctl_cmd(settings, "start", unit="foo.service") == [
        "sudo",
        systemctl,
        "--system",
        "start",
        "foo.service",
    ]


@pytest.mark.usefixtures("systemctl")
def test_systemctl_env(caplog: pytest.LogCaptureFixture) -> None:
    settings = _systemd.Settings(user=False)
    assert systemd.systemctl_env(settings) == {}
    settings = _systemd.Settings(user=True)
    systemd.systemctl_env.cache_clear()
    with mock.patch.dict("os.environ", {}, clear=True), mock.patch(
        "pglift.cmd.run",
        autospec=True,
        side_effect=[
            CompletedProcess(
                [],
                0,
                "/run/user/test\nno\n",
                "",
            ),
            CompletedProcess(
                [],
                0,
                "SOMEVAR=value\nDBUS_SESSION_BUS_ADDRESS=unix:path=/run/user/test/bus\n",
                "",
            ),
        ],
    ) as cmd:
        assert systemd.systemctl_env(settings) == {
            "DBUS_SESSION_BUS_ADDRESS": "unix:path=/run/user/test/bus",
            "XDG_RUNTIME_DIR": "/run/user/test",
        }

    username, _ = default_sysuser()
    assert cmd.mock_calls == [
        mock.call(
            [
                "loginctl",
                "show-user",
                username,
                "--value",
                "--property",
                "RuntimePath",
                "--property",
                "Linger",
            ],
            check=True,
        ),
        mock.call(
            ["systemctl", "--user", "show-environment"],
            env={"XDG_RUNTIME_DIR": "/run/user/test"},
            check=True,
        ),
    ]
    assert [r.message for r in caplog.records] == [
        f"systemd lingering for user {username} is not enabled, "
        "pglift services won't start automatically at boot"
    ]


@pytest.mark.usefixtures("systemctl")
def test_install_uninstall(tmp_path: Path) -> None:
    settings = _systemd.Settings(unit_path=tmp_path)
    systemd.install("foo", "ahah", settings)
    unit_path = tmp_path / "foo"
    mtime = unit_path.stat().st_mtime
    assert unit_path.read_text() == "ahah"
    assert systemd.installed("foo", settings)
    systemd.install("foo", "ahah", settings)
    assert unit_path.stat().st_mtime == mtime
    with pytest.raises(FileExistsError, match="not overwriting"):
        systemd.install("foo", "ahahah", settings)
    assert systemd.installed("foo", settings)
    systemd.uninstall("foo", settings)
    assert not unit_path.exists()
    systemd.uninstall("foo", settings)  # no-op
    assert not systemd.installed("foo", settings)



from pydantic import Field

from .base import BaseModel, LogPath, RunPath


class Settings(BaseModel):
    """Settings for pglift's command-line interface."""

    logpath: LogPath = Field(
        default=LogPath(),
        description="Directory where temporary log files from command executions will be stored",
        title="CLI log directory",
    )

    log_format: str = Field(
        default="%(asctime)s %(levelname)-8s %(name)s - %(message)s",
        description="Format for log messages when written to a file",
    )

    date_format: str = Field(
        default="%Y-%m-%d %H:%M:%S",
        description="Date format in log messages when written to a file",
    )

    lock_file: RunPath = Field(
        default=RunPath(".pglift.lock"),
        description="Path to lock file dedicated to pglift",
    )



from __future__ import annotations

import logging
import shlex
import urllib.parse
from pathlib import Path
from urllib.parse import quote

import psycopg
import psycopg.conninfo
import pydantic
from pgtoolkit.conf import Configuration

from .. import exceptions
from .. import service as service_mod
from ..ctx import Context
from ..models import interface, system
from ..settings import Settings, _prometheus
from ..task import task
from .models import (
    PostgresExporter,
    Service,
    ServiceManifest,
    State,
    default_port,
    service_name,
)

logger = logging.getLogger(__name__)


def available(settings: Settings) -> _prometheus.Settings | None:
    return settings.prometheus


def get_settings(settings: Settings) -> _prometheus.Settings:
    """Return settings for prometheus

    Same as `available` but assert that settings are not None.
    Should be used in a context where settings for the plugin are surely
    set (for example in hookimpl).
    """
    assert settings.prometheus is not None
    return settings.prometheus


def enabled(qualname: str, settings: _prometheus.Settings) -> bool:
    return _configpath(qualname, settings).exists()


def _args(execpath: Path, configpath: Path) -> list[str]:
    varname = "POSTGRES_EXPORTER_OPTS"
    line = config_var(configpath, varname)
    try:
        key, value = line.split("=", 1)
    except ValueError as e:
        raise exceptions.ConfigurationError(
            configpath, f"malformatted {varname} parameter"
        ) from e
    return [str(execpath)] + shlex.split(value[1:-1])


def _configpath(qualname: str, settings: _prometheus.Settings) -> Path:
    return Path(str(settings.configpath).format(name=qualname))


def _env(configpath: Path) -> dict[str, str]:
    env: dict[str, str] = {}
    for line in configpath.read_text().splitlines():
        key, value = line.split("=", 1)
        if key != "POSTGRES_EXPORTER_OPTS":
            env[key] = value
    return env


def _queriespath(qualname: str, settings: _prometheus.Settings) -> Path:
    return Path(str(settings.queriespath).format(name=qualname))


def _pidfile(qualname: str, settings: _prometheus.Settings) -> Path:
    return Path(str(settings.pid_file).format(name=qualname))


def config_var(configpath: Path, varname: str) -> str:
    """Return postgres_exporter configuration file line for given varname.

    :param configpath: the path to the configuration file.
    :param varname: the name of the variable to search for.

    :raises ~exceptions.ConfigurationError: if varname could not be read from
        configuration file.
    :raises ~exceptions.FileNotFoundError: if configuration file is not found.
    """
    if not configpath.exists():
        raise exceptions.FileNotFoundError(
            f"postgres_exporter configuration file {configpath} not found"
        )
    with configpath.open() as f:
        for line in f:
            if line.startswith(varname):
                break
        else:
            raise exceptions.ConfigurationError(configpath, f"{varname} not found")
    return line


def port(name: str, settings: _prometheus.Settings) -> int:
    """Return postgres_exporter port read from configuration file.

    :param name: the name for the service.

    :raises ~exceptions.ConfigurationError: if port could not be read from
        configuration file.
    :raises ~exceptions.FileNotFoundError: if configuration file is not found.
    """
    configpath = _configpath(name, settings)
    varname = "PG_EXPORTER_WEB_LISTEN_ADDRESS"
    line = config_var(configpath, varname)
    try:
        value = line.split("=", 1)[1].split(":", 1)[1]
    except (IndexError, ValueError) as e:
        raise exceptions.ConfigurationError(
            configpath, f"malformatted {varname} parameter"
        ) from e
    return int(value.strip())


def password(name: str, settings: _prometheus.Settings) -> str | None:
    """Return postgres_exporter dsn password read from configuration file.

    :param name: the name for the service.

    :raises ~exceptions.ConfigurationError: if password could not be read from
        configuration file.
    :raises ~exceptions.FileNotFoundError: if configuration file is not found.
    """
    configpath = _configpath(name, settings)
    varname = "DATA_SOURCE_NAME"
    line = config_var(configpath, varname)
    try:
        conninfo = psycopg.conninfo.conninfo_to_dict(line.split("=", 1)[1])
        value: str | None = conninfo.get("password")
    except (IndexError, ValueError, psycopg.ProgrammingError) as e:
        raise exceptions.ConfigurationError(
            configpath, f"malformatted {varname} parameter"
        ) from e
    return value


def make_uri(
    *,
    user: str = "",
    password: str = "",
    port: str = "5432",
    dbname: str = "",
    **kw: str,
) -> str:
    """Return a libpq compatible uri for the given dsn object

    Note: key=value form DSN doesn't work with a unix socket host.
    Also for socket hosts, the host must be given in the uri params
    (after '?').

    >>> make_uri(**{'host': '/socket/path', 'dbname': 'somedb'})
    'postgresql://:5432/somedb?host=%2Fsocket%2Fpath'
    >>> make_uri(**{'host': '/socket/path'})
    'postgresql://:5432?host=%2Fsocket%2Fpath'
    >>> make_uri(**{'host': '/socket/path', 'user': 'someone', 'dbname': 'somedb', 'connect_timeout': '10', 'password': 'secret'})
    'postgresql://someone:secret@:5432/somedb?host=%2Fsocket%2Fpath&connect_timeout=10'
    >>> make_uri(**{'host': '/socket/path', 'user': 'someone', 'dbname': 'somedb', 'password': 'secret@!'})
    'postgresql://someone:secret%40%21@:5432/somedb?host=%2Fsocket%2Fpath'
    """
    userspec = user
    userspec += f":{quote(password)}" if password else ""
    userspec += "@" if userspec else ""
    netloc = f"{userspec}:{port}"
    query = urllib.parse.urlencode(kw)
    return urllib.parse.urlunsplit(("postgresql", netloc, dbname, query, None))


def system_lookup(name: str, settings: _prometheus.Settings) -> Service | None:
    try:
        port_ = port(name, settings)
        passwd_ = password(name, settings)
    except (exceptions.FileNotFoundError, exceptions.ConfigurationError) as exc:
        logger.debug("failed to read postgres_exporter configuration %s: %s", name, exc)
        return None
    else:
        password_ = None
        if passwd_ is not None:
            password_ = pydantic.SecretStr(passwd_)
        return Service(name=name, settings=settings, port=port_, password=password_)


@task
def setup(
    ctx: Context,
    name: str,
    settings: _prometheus.Settings,
    *,
    dsn: str = "",
    password: str | None = None,
    port: int = default_port,
) -> Service:
    """Set up a Prometheus postgres_exporter service for an instance.

    :param name: a (locally unique) name for the service.
    :param dsn: connection info string to target instance.
    :param password: connection password.
    :param port: TCP port for the web interface and telemetry of postgres_exporter.
    """
    uri = make_uri(**psycopg.conninfo.conninfo_to_dict(dsn, password=password))
    config = [f"DATA_SOURCE_NAME={uri}"]
    log_options = ["--log.level=info"]
    opts = " ".join(log_options)
    queriespath = _queriespath(name, settings)
    config.extend(
        [
            f"PG_EXPORTER_WEB_LISTEN_ADDRESS=:{port}",
            f"PG_EXPORTER_EXTEND_QUERY_PATH={queriespath}",
            f"POSTGRES_EXPORTER_OPTS={opts!r}",
        ]
    )

    configpath = _configpath(name, settings)
    needs_restart = False
    if service := system_lookup(name, settings):
        assert configpath.exists()
        if configpath.read_text().splitlines() != config:
            logger.info("reconfiguring Prometheus postgresql %s", name)
            configpath.write_text("\n".join(config))
            needs_restart = True
    else:
        logger.info("configuring Prometheus postgresql %s", name)
        configpath.parent.mkdir(mode=0o750, exist_ok=True, parents=True)
        configpath.touch(mode=0o600)
        configpath.write_text("\n".join(config))
        if not queriespath.exists():
            queriespath.parent.mkdir(mode=0o750, exist_ok=True, parents=True)
            queriespath.touch()
        service = system_lookup(name, settings)
        assert service is not None

    ctx.hook.enable_service(ctx=ctx, service=service_name, name=name)

    if needs_restart:
        restart(ctx, service)
    return service


@setup.revert(title="deconfiguring Prometheus postgres_exporter {name}")
def revert_setup(
    ctx: Context,
    name: str,
    settings: _prometheus.Settings,
    *,
    dsn: str = "",
    password: str | None = None,
    port: int = default_port,
) -> None:
    ctx.hook.disable_service(ctx=ctx, service=service_name, name=name, now=True)
    _configpath(name, settings).unlink(missing_ok=True)
    _queriespath(name, settings).unlink(missing_ok=True)


def start(ctx: Context, service: Service, *, foreground: bool = False) -> None:
    logger.info("starting Prometheus postgres_exporter %s", service.name)
    service_mod.start(ctx, service, foreground=foreground)


def stop(ctx: Context, service: Service) -> None:
    logger.info("stopping Prometheus postgres_exporter %s", service.name)
    service_mod.stop(ctx, service)


def restart(ctx: Context, service: Service) -> None:
    logger.info("restarting Prometheus postgres_exporter %s", service.name)
    service_mod.restart(ctx, service)


def apply(
    ctx: Context,
    postgres_exporter: PostgresExporter,
    settings: _prometheus.Settings,
) -> interface.ApplyResult:
    """Apply state described by specified interface model as a postgres_exporter
    service for a non-local instance.

    :raises exceptions.InstanceStateError: if the target instance exists on system.
    """
    try:
        system.PostgreSQLInstance.from_qualname(postgres_exporter.name, ctx.settings)
    except (ValueError, exceptions.InstanceNotFound):
        pass
    else:
        raise exceptions.InstanceStateError(
            f"instance {postgres_exporter.name!r} exists locally"
        )

    exists = enabled(postgres_exporter.name, settings)
    if postgres_exporter.state == State.absent:
        drop(ctx, postgres_exporter.name)
        return interface.ApplyResult(
            change_state=interface.ApplyChangeState.dropped if exists else None
        )
    else:
        password = None
        if postgres_exporter.password:
            password = postgres_exporter.password.get_secret_value()
        service = setup(
            ctx,
            postgres_exporter.name,
            settings,
            dsn=postgres_exporter.dsn,
            password=password,
            port=postgres_exporter.port,
        )
        if postgres_exporter.state == State.started:
            start(ctx, service)
        elif postgres_exporter.state == State.stopped:
            stop(ctx, service)
        return interface.ApplyResult(
            change_state=interface.ApplyChangeState.created
            if not exists
            else interface.ApplyChangeState.changed
        )


def drop(ctx: Context, name: str) -> None:
    """Remove a postgres_exporter service."""
    logger.info("dropping postgres_exporter service '%s'", name)
    settings = get_settings(ctx.settings)
    if (service := system_lookup(name, settings)) is None:
        logger.warning("no postgres_exporter service '%s' found", name)
        return
    stop(ctx, service)
    revert_setup(ctx, name, settings)


def setup_local(
    ctx: Context,
    manifest: interface.Instance,
    settings: _prometheus.Settings,
    instance_config: Configuration,
) -> None:
    """Setup Prometheus postgres_exporter for a local instance."""
    service_manifest = manifest.service_manifest(ServiceManifest)
    rolename = settings.role
    dsn = ["dbname=postgres"]
    if "port" in instance_config:
        dsn.append(f"port={instance_config.port}")
    if host := instance_config.get("unix_socket_directories"):
        dsn.append(f"host={host}")
    dsn.append(f"user={rolename}")
    if not instance_config.get("ssl", False):
        dsn.append("sslmode=disable")

    instance = system.PostgreSQLInstance.system_lookup(
        (manifest.name, manifest.version, ctx.settings)
    )
    configpath = _configpath(instance.qualname, settings)
    password_: str | None = None
    if service_manifest.password:
        password_ = service_manifest.password.get_secret_value()
    elif configpath.exists():
        # Get the password from config file
        password_ = password(instance.qualname, settings)

    setup(
        ctx,
        instance.qualname,
        settings,
        dsn=" ".join(dsn),
        password=password_,
        port=service_manifest.port,
    )



from __future__ import annotations

import configparser
import contextlib
import datetime
import json
import logging
import os
import re
from collections.abc import Iterator
from functools import partial
from pathlib import Path
from typing import Any

from dateutil.tz import gettz
from pgtoolkit import conf as pgconf

from .. import cmd, exceptions, types, util
from ..ctx import Context
from ..models import interface, system
from ..postgresql.ctl import libpq_environ
from ..settings import Settings, _pgbackrest
from ..task import task
from .models import Backup, Service, ServiceManifest

logger = logging.getLogger(__name__)


def available(settings: Settings) -> _pgbackrest.Settings | None:
    return settings.pgbackrest


def get_settings(settings: Settings) -> _pgbackrest.Settings:
    """Return settings for pgbackrest

    Same as `available` but assert that settings are not None.
    Should be used in a context where settings for the plugin are surely
    set (for example in hookimpl).
    """
    assert settings.pgbackrest is not None
    return settings.pgbackrest


def enabled(
    instance: system.PostgreSQLInstance, settings: _pgbackrest.Settings
) -> bool:
    return system_lookup(instance.datadir, settings) is not None


def base_configpath(settings: _pgbackrest.Settings) -> Path:
    return settings.configpath / "pgbackrest.conf"


def config_directory(settings: _pgbackrest.Settings) -> Path:
    return settings.configpath / "conf.d"


def create_include_directory(settings: _pgbackrest.Settings) -> None:
    logger.info("creating pgbackrest include directory")
    config_directory(settings).mkdir(mode=0o750, exist_ok=True, parents=True)


def delete_include_directory(settings: _pgbackrest.Settings) -> None:
    logger.info("deleting pgbackrest include directory")
    util.rmdir(config_directory(settings))


def make_cmd(stanza: str, settings: _pgbackrest.Settings, *args: str) -> list[str]:
    return [
        str(settings.execpath),
        f"--config-path={settings.configpath}",
        "--log-level-stderr=info",
        f"--stanza={stanza}",
    ] + list(args)


parser = partial(configparser.ConfigParser, strict=True)

pgpath_rgx = re.compile(r"pg(\d+)-path")


def next_index(cp: configparser.ConfigParser, stanza: str) -> int:
    """Return the next index of pgN- options in parser.

    >>> cp = parser()
    >>> next_index(cp, "st")
    Traceback (most recent call last):
        ...
    configparser.NoSectionError: No section: 'st'
    >>> cp.add_section("st")
    >>> next_index(cp, "st")
    1
    >>> cp["st"]["pg1-path"] = "/pgsql/main"
    >>> next_index(cp, "st")
    2
    """
    idxs = {0}
    for opt in cp.options(stanza):
        if m := pgpath_rgx.match(opt):
            idxs.add(int(m.group(1)))
    return max(idxs) + 1


def system_lookup(datadir: Path, settings: _pgbackrest.Settings) -> Service | None:
    d = config_directory(settings)
    for p in d.glob("*.conf"):
        cp = parser()
        with p.open() as f:
            cp.read_file(f)
        for stanza in cp.sections():
            for key, value in cp.items(stanza):
                if (m := pgpath_rgx.match(key)) and value == str(datadir):
                    return Service(stanza=stanza, path=p, index=int(m.group(1)))
    return None


def stanza_pgpaths(path: Path, stanza: str) -> Iterator[tuple[int, Path]]:
    cp = parser()
    with path.open() as f:
        cp.read_file(f)
    for key, value in cp.items(stanza):
        if m := pgpath_rgx.match(key):
            yield int(m.group(1)), Path(value)


def service(
    instance: system.BaseInstance,
    manifest: ServiceManifest,
    settings: _pgbackrest.Settings,
    upgrading_from: system.Instance | None,
) -> Service:
    """Retrieve a Service object.

    In case of an upgrade, if a service exists on original 'upgrading_from'
    instance, return it.

    Otherwise, lookup for a service bound to 'instance' and return it if
    found.

    Fall back to building a new service object from scratch.
    """
    if upgrading_from:
        if s := system_lookup(upgrading_from.datadir, settings):
            return s
    if s := system_lookup(instance.datadir, settings):
        if manifest.stanza != s.stanza:
            raise exceptions.InstanceStateError(
                f"instance {instance} is already bound to pgbackrest stanza {s.stanza!r} (path={s.path})"
            )
        return s
    stanza = manifest.stanza
    index = 1
    path = config_directory(settings) / f"{stanza}.conf"
    # Use the next pgN index when the path exists but configured instance is
    # not attached to its stanza. This would typically happen for standby
    # instances, when created on the same host as the primary.
    if path.exists():
        cp = parser()
        with path.open() as f:
            cp.read_file(f)
        index = next_index(cp, stanza)
    return Service(stanza=stanza, path=path, index=index)


def backup_info(
    service: Service,
    settings: _pgbackrest.Settings,
    *,
    backup_set: str | None = None,
) -> dict[str, Any]:
    """Call pgbackrest info command to obtain information about backups.

    Ref.: https://pgbackrest.org/command.html#command-info
    """
    args = []
    if backup_set is not None:
        args.append(f"--set={backup_set}")
    args.extend(["--output=json", "info"])
    r = cmd.run(make_cmd(service.stanza, settings, *args), check=True)
    infos = json.loads(r.stdout)
    try:
        return next(i for i in infos if i["name"] == service.stanza)
    except StopIteration:
        return {}


@task
def setup(
    ctx: Context,
    service: Service,
    settings: _pgbackrest.Settings,
    instance_config: pgconf.Configuration,
    changes: types.ConfigChanges,
    creating: bool,
    datadir: Path,
) -> None:
    """Setup pgBackRest"""
    if not creating and not any(
        s in changes for s in ("port", "unix_socket_directories")
    ):
        return

    base_config_path = base_configpath(settings)
    if not base_config_path.exists():
        raise exceptions.SystemError(
            f"Missing base config file {base_config_path} for pgbackrest. "
            "Did you forget to run 'pglift site-configure'?"
        )

    stanza = service.stanza
    stanza_confpath = service.path
    pg = f"pg{service.index}"
    cp = parser()
    if stanza_confpath.exists():
        with stanza_confpath.open() as f:
            cp.read_file(f)

    logger.info(
        "configuring pgBackRest stanza '%s' for pg%d-path=%s",
        stanza,
        service.index,
        datadir,
    )
    # Always use string values so that this would match with actual config (on
    # disk) that's parsed later on.
    config = {
        stanza: {
            f"{pg}-path": str(datadir),
            f"{pg}-port": str(instance_config.get("port", 5432)),
            f"{pg}-user": ctx.settings.postgresql.backuprole.name,
        },
    }
    if unix_socket_directories := instance_config.get("unix_socket_directories"):
        config[stanza][f"{pg}-socket-path"] = str(unix_socket_directories)
    cp.read_dict(config)
    with stanza_confpath.open("w") as configfile:
        cp.write(configfile)


def postgresql_configuration(
    stanza: str, settings: _pgbackrest.Settings, datadir: Path
) -> pgconf.Configuration:
    pgconfig = util.template("postgresql", "pgbackrest.conf").format(
        execpath=settings.execpath,
        configpath=settings.configpath,
        stanza=stanza,
        datadir=datadir,
    )
    config = pgconf.Configuration()
    list(config.parse(pgconfig.splitlines()))
    return config


@setup.revert(title="deconfiguring pgBackRest")
def revert_setup(
    ctx: Context,
    service: Service,
    settings: _pgbackrest.Settings,
    instance_config: pgconf.Configuration,
    changes: types.ConfigChanges,
    creating: bool,
    datadir: Path,
) -> None:
    """Un-setup pgBackRest.

    Remove options from 'stanza' section referencing instance's datadir, then
    possibly remove the configuration file if empty, and finally remove
    stanza's log files.
    """
    stanza = service.stanza
    stanza_confpath = service.path
    if stanza_confpath.exists():
        cp = parser()
        with stanza_confpath.open() as f:
            cp.read_file(f)
        for opt in cp.options(stanza):
            if opt.startswith(f"pg{service.index}-"):
                cp.remove_option(stanza, opt)
        with stanza_confpath.open("w") as f:
            cp.write(f)
        if not cp.options(stanza):
            stanza_confpath.unlink(missing_ok=True)
    if not stanza_confpath.exists():
        for logf in settings.logpath.glob(f"{stanza}-*.log"):
            logf.unlink(missing_ok=True)


def check(
    instance: system.PostgreSQLInstance,
    service: Service,
    settings: _pgbackrest.Settings,
    password: str | None,
) -> None:
    env = os.environ.copy()
    if password is not None:
        env["PGPASSWORD"] = password
    postgresql_settings = instance._settings.postgresql
    env = libpq_environ(instance, postgresql_settings.backuprole.name, base=env)
    cmd.run(make_cmd(service.stanza, settings, "check"), check=True, env=env)


def iter_backups(
    instance: system.Instance, settings: _pgbackrest.Settings
) -> Iterator[Backup]:
    """Yield information about backups on an instance."""
    service = instance.service(Service)
    backups = backup_info(service, settings)["backup"]

    def started_at(entry: Any) -> float:
        return entry["timestamp"]["start"]  # type: ignore[no-any-return]

    for backup in sorted(backups, key=started_at, reverse=True):
        info_set = backup_info(service, settings, backup_set=backup["label"])
        databases = [db["name"] for db in info_set["backup"][0]["database-ref"]]
        dtstart = datetime.datetime.fromtimestamp(backup["timestamp"]["start"])
        dtstop = datetime.datetime.fromtimestamp(backup["timestamp"]["stop"])
        yield Backup(
            label=backup["label"],
            size=backup["info"]["size"],
            repo_size=backup["info"]["repository"]["size"],
            date_start=dtstart.replace(tzinfo=gettz()),
            date_stop=dtstop.replace(tzinfo=gettz()),
            type=backup["type"],
            databases=databases,
        )


def restore_command(
    instance: system.Instance,
    settings: _pgbackrest.Settings,
    *,
    date: datetime.datetime | None = None,
    backup_set: str | None = None,
) -> list[str]:
    """Return the pgbackrest restore for ``instance``.

    Ref.: https://pgbackrest.org/command.html#command-restore
    """
    args = [
        # The delta option allows pgBackRest to handle instance data/wal
        # directories itself, without the need to clean them up beforehand.
        "--delta",
        "--link-all",
    ]
    if date is not None and backup_set is not None:
        raise exceptions.UnsupportedError(
            "date and backup_set are not expected to be both specified"
        )
    elif date is not None:
        target = date.strftime("%Y-%m-%d %H:%M:%S.%f%z")
        args += ["--target-action=promote", "--type=time", f"--target={target}"]
    elif backup_set is not None:
        args += ["--target-action=promote", "--type=immediate", f"--set={backup_set}"]
    args.append("restore")
    s = instance.service(Service)
    return make_cmd(s.stanza, settings, *args)


def restore(
    ctx: Context,
    instance: system.Instance,
    settings: _pgbackrest.Settings,
    *,
    label: str | None = None,
    date: datetime.datetime | None = None,
) -> None:
    """Restore an instance, possibly only including specified databases.

    The instance must not be running.

    Ref.: https://pgbackrest.org/command.html#command-restore
    """
    logger.info("restoring instance %s with pgBackRest", instance)
    if instance.standby:
        raise exceptions.InstanceReadOnlyError(instance)

    cmd_args = restore_command(instance, settings, date=date, backup_set=label)
    cmd.run(cmd_args, check=True)


def env_for(service: Service, settings: _pgbackrest.Settings) -> dict[str, str]:
    return {
        "PGBACKREST_CONFIG_PATH": str(settings.configpath),
        "PGBACKREST_STANZA": service.stanza,
    }


def site_configure_install(
    settings: Settings,
    config: configparser.ConfigParser,
) -> None:
    s = get_settings(settings)
    logger.info("installing base pgbackrest configuration")
    global_configpath = base_configpath(s)
    global_configpath.parent.mkdir(parents=True, exist_ok=True)
    with global_configpath.open("w") as f:
        config.write(f)
    create_include_directory(s)


def site_configure_uninstall(settings: Settings) -> None:
    s = get_settings(settings)
    delete_include_directory(s)
    logger.info("uninstalling base pgbackrest configuration")
    base_configpath(s).unlink(missing_ok=True)


@contextlib.contextmanager
def instance_configured(
    ctx: Context,
    manifest: interface.Instance,
    config: pgconf.Configuration,
    changes: types.ConfigChanges,
    creating: bool,
    upgrading_from: system.Instance | None,
) -> Iterator[system.PostgreSQLInstance]:
    instance = system.PostgreSQLInstance.system_lookup(
        (manifest.name, manifest.version, ctx.settings)
    )
    yield instance
    settings = get_settings(instance._settings)
    service_manifest = manifest.service_manifest(ServiceManifest)
    service_ = service(instance, service_manifest, settings, upgrading_from)
    setup(ctx, service_, settings, config, changes, creating, instance.datadir)


@contextlib.contextmanager
def instance_dropped(
    ctx: Context, instance: system.Instance
) -> Iterator[Service | None]:
    try:
        service = instance.service(Service)
    except ValueError:
        yield None
        return
    settings = get_settings(instance._settings)
    yield service
    revert_setup(ctx, service, settings, instance.config(), {}, False, instance.datadir)



from __future__ import annotations

import asyncio
import asyncio.subprocess
import logging
import os
import shlex
import signal
import subprocess
from asyncio import create_task
from collections.abc import AsyncIterator, Mapping, Sequence
from contextlib import asynccontextmanager
from functools import partial
from pathlib import Path
from subprocess import PIPE, TimeoutExpired
from types import TracebackType
from typing import IO, Any, NoReturn

from . import exceptions
from ._compat import Self
from .types import CompletedProcess, Status

logger = logging.getLogger(__name__)


class _CloneStderrProtocol(asyncio.subprocess.SubprocessStreamProtocol):
    """Subprocess protocol extending the default one to handle a clone of stderr stream."""

    def __init__(
        self,
        stderr_reader: asyncio.StreamReader | None,
        *,
        limit: int,
        loop: asyncio.events.AbstractEventLoop,
    ) -> None:
        super().__init__(limit=limit, loop=loop)
        self._stderr_reader = stderr_reader

    def __repr__(self) -> str:
        base = super().__repr__()[1:-1]
        if self._stderr_reader:
            base += f" stderr(clone)={self._stderr_reader}"
        return f"<{base}>"

    def pipe_data_received(self, fd: int, data: bytes | str) -> None:
        super().pipe_data_received(fd, data)
        if fd == 2 and self._stderr_reader:
            assert isinstance(data, bytes)
            self._stderr_reader.feed_data(data)

    def pipe_connection_lost(self, fd: int, exc: Exception | None) -> None:
        super().pipe_connection_lost(fd, exc)
        if fd == 2 and self._stderr_reader:
            if exc:
                self._stderr_reader.set_exception(exc)
            else:
                self._stderr_reader.feed_eof()


@asynccontextmanager
async def logged_subprocess_exec(
    program: str,
    *args: str,
    stdin: int | IO[Any] | None = None,
    stdout: int | IO[Any] | None = None,
    stderr: int | IO[Any] | None = None,
    **kwds: Any,
) -> AsyncIterator[asyncio.subprocess.Process]:
    """Context manager starting an asyncio Process while possibly processing its
    stderr stream with 'stderr_handler' callback.

    This is similar quite to asyncio.subprocess.create_subprocess_exec() but
    with a custom protocol to install a cloned stream for stderr.
    """
    loop = asyncio.get_event_loop()
    task = None
    cloned_stderr = None
    if stderr is not None:
        cloned_stderr = asyncio.StreamReader()

        async def handle_stderr(stream: asyncio.StreamReader) -> None:
            async for line in stream:
                logger.debug("%s: %s", program, line.decode("utf-8").rstrip())

        task = create_task(handle_stderr(cloned_stderr), name="stderr logger")

    protocol_factory = partial(
        _CloneStderrProtocol,
        cloned_stderr,
        limit=2**16,  # asyncio.streams._DEFAULT_LIMIT
        loop=loop,
    )
    try:
        transport, protocol = await loop.subprocess_exec(
            protocol_factory,
            program,
            *args,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            **kwds,
        )
        yield asyncio.subprocess.Process(transport, protocol, loop)
    finally:
        if task and not task.done():
            task.cancel()


def run(
    args: Sequence[str],
    *,
    input: str | None = None,
    check: bool = False,
    timeout: float | None = None,
    **kwargs: Any,
) -> CompletedProcess:
    """Run a command as a subprocess while forwarning its stderr to module 'logger'.

    Standard output and errors of child subprocess are captured by default.

    >>> run(["true"], input="a", capture_output=False)
    CompletedProcess(args=['true'], returncode=0)

    Files can also be used with ``stdout`` and ``stderr`` arguments:

    >>> import tempfile
    >>> with tempfile.NamedTemporaryFile() as f:
    ...     _ = run(["echo", "ahah"], stdout=f, stderr=None)
    ...     with open(f.name) as f:
    ...         print(f.read(), end="")
    ahah

    >>> r = run(["cat", "doesnotexist"], stdout=PIPE, stderr=PIPE, env={"LANG": "C"})
    >>> print(r.stderr, end="")
    cat: doesnotexist: No such file or directory

    With ``check=True``, :class:`~pglift.exceptions.CommandError` is raised
    in case of non-zero return code:

    >>> run(["cat", "doesnotexist"], check=True)
    Traceback (most recent call last):
        ...
    pglift.exceptions.CommandError: Command '['cat', 'doesnotexist']' returned non-zero exit status 1.

    With a non-None timeout argument, a :class:`~subprocess.TimeoutExpired`
    exception might be raised:

    >>> run(["sleep", "0.1"], timeout=0.01)
    Traceback (most recent call last):
        ...
    subprocess.TimeoutExpired: Command '['sleep', '0.1']' timed out after 0.01 seconds

    >>> run(["nosuchcommand", "x", "y", "-v"])
    Traceback (most recent call last):
        ...
    pglift.exceptions.FileNotFoundError: program from command 'nosuchcommand x y -v' not found
    """
    cmds = shlex.join(args)
    logger.debug(cmds)
    if not args:
        raise ValueError("empty arguments sequence")

    if input is not None:
        if "stdin" in kwargs:
            raise ValueError("stdin and input arguments may not both be used")
        kwargs["stdin"] = PIPE

    try:
        capture_output = kwargs.pop("capture_output")
    except KeyError:
        kwargs.setdefault("stdout", subprocess.PIPE)
        kwargs.setdefault("stderr", subprocess.PIPE)
    else:
        if capture_output:
            if "stdout" in kwargs or "stderr" in kwargs:
                raise ValueError(
                    "stdout and stderr arguments may not be used with capture_output"
                )
            kwargs["stdout"] = kwargs["stderr"] = subprocess.PIPE

    loop = asyncio.get_event_loop()

    async def run() -> tuple[asyncio.subprocess.Process, str | None, str | None]:
        async with logged_subprocess_exec(*args, **kwargs) as proc:
            aw = proc.communicate(input.encode("utf-8") if input is not None else None)
            if timeout is None:
                out, err = await aw
            else:
                try:
                    out, err = await asyncio.wait_for(aw, timeout)
                except asyncio.TimeoutError:
                    raise TimeoutExpired(args, timeout) from None
        assert proc.returncode is not None
        return (
            proc,
            out.decode("utf-8") if out is not None else None,
            err.decode("utf-8") if err is not None else None,
        )

    try:
        proc, out, err = loop.run_until_complete(run())
    except FileNotFoundError as e:
        raise exceptions.FileNotFoundError(
            f"program from command {cmds!r} not found"
        ) from e
    except OSError as e:
        logger.debug("failed to start child process", exc_info=True)
        raise exceptions.SystemError(
            f"failed to start child process from command {cmds!r}"
        ) from e

    if check and proc.returncode:
        raise exceptions.CommandError(proc.returncode, args, out, err)

    assert proc.returncode is not None
    return CompletedProcess(args, proc.returncode, out, err)


def execute_program(
    cmd: Sequence[str], *, env: Mapping[str, str] | None = None
) -> NoReturn:
    """Execute program described by 'cmd', replacing the current process.

    :raises ValueError: if program path is not absolute.
    """
    program = cmd[0]
    if not Path(program).is_absolute():
        raise ValueError(f"expecting an absolute program path {program}")
    logger.debug("executing program '%s'", shlex.join(cmd))
    if env is not None:
        os.execve(program, list(cmd), env)  # nosec
    else:
        os.execv(program, list(cmd))  # nosec


def status_program(pidfile: Path) -> Status:
    """Return the status of a program which PID is in 'pidfile'.

    :raises ~exceptions.SystemError: if the program is already running.
    :raises ~exceptions.CommandError: in case program execution terminates
        after `timeout`.
    """
    if pidfile.exists():
        with pidfile.open() as f:
            pid = f.readline().rstrip()
        if (Path("/proc") / pid).exists():
            return Status.running
    return Status.not_running


class Program:
    """Start program described by 'cmd' and possibly store its PID in 'pidfile'.

    This is aimed at starting daemon programs.

    :raises ~exceptions.SystemError: if the program is already running.
    :raises ~exceptions.CommandError: in case program execution terminates
        after `timeout`.

    When used as a context manager, any exception raised within the block will
    trigger program termination at exit. This can be used to perform sanity
    checks shortly after program startup.
    """

    def __init__(
        self,
        cmd: Sequence[str],
        pidfile: Path | None,
        *,
        timeout: float = 1,
        env: Mapping[str, str] | None = None,
        capture_output: bool = True,
    ) -> None:
        self.pidfile = pidfile
        self.program = cmd[0]
        self.args = cmd
        self.proc = self._start(
            cmd, timeout=timeout, env=env, capture_output=capture_output
        )

    def _start(
        self,
        cmd: Sequence[str],
        *,
        timeout: float = 1,
        env: Mapping[str, str] | None = None,
        capture_output: bool = True,
    ) -> asyncio.subprocess.Process:
        if self.pidfile is not None:
            self._check_pidfile()
        stdout = stderr = None
        if capture_output:
            stdout = stderr = subprocess.PIPE
        logger.debug("starting program '%s'", shlex.join(cmd))

        async def run() -> asyncio.subprocess.Process:
            async with logged_subprocess_exec(
                *cmd, stdout=stdout, stderr=stderr, env=env
            ) as proc:
                aw = proc.communicate()
                try:
                    __, errs = await asyncio.wait_for(aw, timeout)
                except asyncio.TimeoutError:
                    if self.pidfile is not None:
                        self.pidfile.parent.mkdir(parents=True, exist_ok=True)
                        self.pidfile.write_text(str(proc.pid))
                    return proc
                else:
                    assert proc.returncode is not None
                    assert (
                        proc.returncode != 0
                    ), f"{self.program} terminated with exit code 0"
                    raise exceptions.CommandError(
                        proc.returncode,
                        cmd,
                        stderr=errs.decode("utf-8") if errs is not None else None,
                    )

        loop = asyncio.get_event_loop()
        return loop.run_until_complete(run())

    def _check_pidfile(self) -> None:
        """Use specified pidfile, when not None, to check if the program is
        already running.
        """
        pidfile = self.pidfile
        assert pidfile is not None
        if (status := status_program(pidfile)) is Status.running:
            with pidfile.open() as f:
                pid = f.readline().strip()
            if status == Status.running:
                raise exceptions.SystemError(
                    f"program {self.program} seems to be running already with PID {pid}"
                )
        elif pidfile.exists():
            with pidfile.open() as f:
                pid = f.readline().strip()
            logger.warning(
                "program %s is supposed to be running with PID %s but "
                "it's apparently not; starting anyway",
                self.program,
                pid,
            )
            pidfile.unlink()

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException,
        traceback: TracebackType | None,
    ) -> None:
        if exc_value is not None:
            cmd = shlex.join(self.args)
            logger.warning("terminating program '%s'", cmd)
            try:
                self.proc.terminate()
            except ProcessLookupError:
                logger.debug("program %s already terminated", self.args[0])
            if self.proc.stderr is not None:

                async def exit(stderr: asyncio.StreamReader) -> None:
                    async for line in stderr:
                        logger.debug(
                            "%s: %s", self.program, line.decode("utf-8").rstrip()
                        )

                loop = asyncio.get_event_loop()
                loop.run_until_complete(exit(self.proc.stderr))
            if self.pidfile is not None:
                self.pidfile.unlink()
            raise exc_value


def terminate_program(pidfile: Path) -> None:
    """Terminate program matching PID in 'pidfile'.

    Upon successful termination, the 'pidfile' is removed.
    No-op if no process matching PID from 'pidfile' is running.
    """
    if status_program(pidfile) == Status.not_running:
        logger.warning("program from %s not running", pidfile)
        if pidfile.exists():
            logger.debug("removing dangling PID file %s", pidfile)
            pidfile.unlink()
        return

    with pidfile.open() as f:
        pid = int(f.readline().rstrip())
    logger.debug("terminating process %d", pid)
    try:
        os.kill(pid, signal.SIGTERM)
    except ProcessLookupError as e:
        logger.warning("failed to kill process %d: %s", pid, e)
    pidfile.unlink()


def _main() -> None:
    import argparse
    import logging
    import sys

    logger = logging.getLogger("stderr")
    logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(
        logging.Formatter(fmt="%(asctime)s - %(message)s", datefmt="[%Xs]")
    )
    logger.addHandler(handler)

    parser = argparse.ArgumentParser(
        __name__,
        description="Run, start or terminate programs while logging their stderr",
    )
    subparsers = parser.add_subparsers(title="Commands")

    run_parser = subparsers.add_parser(
        "run",
        description="Run PROGRAM with positional ARGuments.",
        epilog=f"Example: {__name__} run initdb /tmp/pgdata --debug",
    )
    run_parser.add_argument("program", metavar="PROGRAM")
    run_parser.add_argument("arguments", metavar="ARG", nargs="*")

    def run_func(args: argparse.Namespace, remaining: Sequence[str]) -> None:
        cmd = [args.program] + args.arguments + list(remaining)
        run(cmd, check=True, logger=logger)

    run_parser.set_defaults(func=run_func)

    start_parser = subparsers.add_parser(
        "start",
        description="Start PROGRAM with positional ARGuments.",
        epilog=f"Example: {__name__} start postgres -D /tmp/pgdata -k /tmp",
    )
    start_parser.add_argument("program", metavar="PROGRAM")
    start_parser.add_argument("arguments", metavar="ARG", nargs="*")
    start_parser.add_argument(
        "-p",
        "--pidfile",
        type=Path,
        help="Path to file where PID will be stored.",
    )
    start_parser.add_argument(
        "--timeout", type=float, default=1, help="Liveliness timeout."
    )

    def start_func(args: argparse.Namespace, remaining: Sequence[str]) -> None:
        cmd = [args.program] + args.arguments + list(remaining)
        with Program(cmd, pidfile=args.pidfile, timeout=args.timeout) as p:
            print(f"Program {args.program} running with PID {p.proc.pid}")

    start_parser.set_defaults(func=start_func)

    terminate_parser = subparsers.add_parser("terminate")
    terminate_parser = subparsers.add_parser(
        "terminate",
        description="Terminate process from PIDFILE.",
        epilog=f"Example: {__name__} terminate /tmp/pgdata/postmaster.pid",
    )
    terminate_parser.add_argument("pidfile", metavar="PIDFILE", type=Path)

    def terminate_func(args: argparse.Namespace, remaining: Sequence[str]) -> None:
        terminate_program(args.pidfile)

    terminate_parser.set_defaults(func=terminate_func)

    ns, remaining = parser.parse_known_args()
    ns.func(ns, remaining)


if __name__ == "__main__":
    _main()



from __future__ import annotations

import collections
import contextlib
import functools
import inspect
import logging
from collections.abc import Iterator
from dataclasses import dataclass
from typing import Any, Callable, ClassVar, Deque, Generic, Protocol, TypeVar, overload

import pydantic
from typing_extensions import ParamSpec

from . import __name__ as pkgname
from . import exceptions

P = ParamSpec("P")
T = TypeVar("T")

Call = tuple["Task[Any, Any]", tuple[Any, ...], dict[str, Any]]

logger = logging.getLogger(pkgname)


class Displayer(Protocol):
    def handle(self, msg: str) -> None:
        ...


@dataclass
class RevertAction(Generic[P]):
    signature: inspect.Signature
    call: Callable[P, Any]


class Task(Generic[P, T]):
    _calls: ClassVar[Deque[Call] | None] = None
    displayer: ClassVar[Displayer | None] = None

    def __init__(self, title: str | None, action: Callable[P, T]) -> None:
        self.title = title
        self.action = action
        self.signature = inspect.signature(action)
        self.revert_action: RevertAction[P] | None = None
        functools.update_wrapper(self, action)

    def __repr__(self) -> str:
        return f"<Task {self.action.__name__!r} at 0x{id(self)}>"

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        if self._calls is not None:
            self._calls.append((self, args, kwargs))
        b = self.signature.bind(*args, **kwargs)
        b.apply_defaults()
        if self.title:
            self.display(self.title.format(**b.arguments))
        return self.action(*args, **kwargs)

    def display(self, title: str) -> None:
        if self.displayer is not None:
            self.displayer.handle(title)

    @overload
    def revert(self, __func: Callable[P, Any]) -> Callable[P, Any]:
        ...

    @overload
    def revert(
        self, *, title: str | None = None
    ) -> Callable[[Callable[P, Any]], Callable[P, Any]]:
        ...

    def revert(
        self, __func: Callable[P, Any] | None = None, *, title: str | None = None
    ) -> Callable[P, Any] | Callable[[Callable[P, Any]], Callable[P, Any]]:
        """Decorator to register a 'revert' callback function.

        The revert function must accept the same arguments than its respective
        action.
        """

        def decorator(revertfn: Callable[P, Any]) -> Callable[P, Any]:
            s = inspect.signature(revertfn)
            assert s.parameters == self.signature.parameters, (
                f"Parameters of function {self.action.__module__}.{self.action.__name__}({self.signature}) "
                f"differ from related revert function {revertfn.__module__}.{revertfn.__name__}({s})"
            )

            @functools.wraps(revertfn)
            def wrapper(*args: P.args, **kwargs: P.kwargs) -> Any:
                b = s.bind(*args, **kwargs)
                b.apply_defaults()
                if title is not None:
                    self.display(title.format(**b.arguments))
                return revertfn(*args, **kwargs)

            self.revert_action = RevertAction(s, wrapper)
            return wrapper

        if __func is not None:
            return decorator(__func)
        return decorator

    def rollback(self, *args: P.args, **kwargs: P.kwargs) -> None:
        if self.revert_action is None:
            return
        if self.title:
            b = self.revert_action.signature.bind(*args, **kwargs)
            b.apply_defaults()
            logger.warning("reverting: %s", self.title.format(**b.arguments))
        self.revert_action.call(*args, **kwargs)


@overload
def task(__func: Callable[P, T]) -> Task[P, T]:
    ...


@overload
def task(*, title: str | None) -> Callable[[Callable[P, T]], Task[P, T]]:
    ...


def task(
    __func: Callable[P, T] | None = None, *, title: str | None = None
) -> Task[P, T] | Callable[[Callable[P, T]], Task[P, T]]:
    def mktask(fn: Callable[P, T]) -> Task[P, T]:
        return functools.wraps(fn)(Task(title, fn))

    if __func is not None:
        return mktask(__func)
    return mktask


@contextlib.contextmanager
def displayer_installed(displayer: Displayer | None) -> Iterator[None]:
    if displayer is None:
        yield
        return
    assert Task.displayer is None
    Task.displayer = displayer
    try:
        yield
    finally:
        Task.displayer = None


@contextlib.contextmanager
def transaction(revert_on_error: bool = True) -> Iterator[None]:
    """Context manager handling revert of run tasks, in case of failure."""
    if Task._calls is not None:
        raise RuntimeError("inconsistent task state")
    Task._calls = collections.deque()
    try:
        yield
    except BaseException as exc:
        # Only log internal errors, i.e. those not coming from user
        # cancellation or invalid input data.
        if isinstance(exc, KeyboardInterrupt):
            if Task._calls:
                logger.warning("%s interrupted", Task._calls[-1][0])
        elif not isinstance(exc, (pydantic.ValidationError, exceptions.Cancelled)):
            logger.warning(str(exc))
        assert Task._calls is not None
        while True:
            try:
                t, args, kwargs = Task._calls.pop()
            except IndexError:
                break
            if revert_on_error:
                t.rollback(*args, **kwargs)
        raise exc
    finally:
        Task._calls = None



from __future__ import annotations

import sys

from pglift.backup import systemd_unit_templates
from pglift.settings import Settings


def test_systemd_unit_templates(settings: Settings) -> None:
    ((service_name, service_content), (timer_name, timer_content)) = list(
        systemd_unit_templates(settings, env={"X-DEBUG": "no"}, content=True)
    )
    assert service_name == "pglift-backup@.service"
    service_lines = service_content.splitlines()
    for line in service_lines:
        if line.startswith("ExecStart"):
            execstart = line.split("=", 1)[-1]
            assert execstart == f"{sys.executable} -m pglift instance backup %I"
            break
    else:
        raise AssertionError("ExecStart line not found")
    assert 'Environment="X-DEBUG=no"' in service_lines
    assert timer_name == "pglift-backup@.timer"
    timer_lines = timer_content.splitlines()
    assert "OnCalendar=daily" in timer_lines



from __future__ import annotations

import configparser
import logging

import pgtoolkit.conf as pgconf

from .. import hookimpl, types, util
from ..ctx import Context
from ..models import interface, system
from ..settings import Settings, _pgbackrest
from . import base
from . import register_if as base_register_if
from .base import get_settings, parser

HostRepository = _pgbackrest.SSHHostRepository
logger = logging.getLogger(__name__)


def register_if(settings: Settings) -> bool:
    if not base_register_if(settings):
        return False
    s = get_settings(settings)
    return isinstance(s.repository, HostRepository)


@hookimpl
def site_configure_install(settings: Settings) -> None:
    s = get_settings(settings)
    base.site_configure_install(settings, base_config(s))


@hookimpl
def site_configure_uninstall(settings: Settings) -> None:
    base.site_configure_uninstall(settings)


@hookimpl
def instance_configured(
    ctx: Context,
    manifest: interface.Instance,
    config: pgconf.Configuration,
    changes: types.ConfigChanges,
    creating: bool,
    upgrading_from: system.Instance | None,
) -> None:
    with base.instance_configured(
        ctx, manifest, config, changes, creating, upgrading_from
    ):
        pass


@hookimpl
def instance_dropped(ctx: Context, instance: system.Instance) -> None:
    with base.instance_dropped(ctx, instance):
        pass


def repository_settings(settings: _pgbackrest.Settings) -> HostRepository:
    assert isinstance(settings.repository, HostRepository)
    return settings.repository


def base_config(settings: _pgbackrest.Settings) -> configparser.ConfigParser:
    """Build the base configuration for pgbackrest clients on the database
    host.
    """
    cp = parser()
    cp.read_string(
        util.template("pgbackrest", "pgbackrest.conf").format(**dict(settings))
    )
    s = repository_settings(settings)
    rhost = {
        "repo1-host-type": "ssh",
        "repo1-host": s.host,
    }
    if s.host_port:
        rhost["repo1-host-port"] = str(s.host_port)
    if s.host_config:
        rhost["repo1-host-config"] = str(s.host_config)
    if s.host_user:
        rhost["repo1-host-user"] = s.host_user
    if s.cmd_ssh:
        rhost["cmd-ssh"] = str(s.cmd_ssh)
    cp["global"].update(rhost)
    return cp



from __future__ import annotations

import logging
from typing import Any

from pgtoolkit.conf import Configuration
from pydantic import Field

from .. import hookimpl
from ..models import interface
from ..settings import Settings
from . import models
from .impl import POWA_EXTENSIONS, POWA_LIBRARIES
from .impl import available as available

logger = logging.getLogger(__name__)


def register_if(settings: Settings) -> bool:
    return available(settings) is not None


@hookimpl
def instance_settings() -> Configuration:
    conf = Configuration()
    conf["shared_preload_libraries"] = ", ".join(POWA_LIBRARIES)
    return conf


@hookimpl
def interface_model() -> tuple[str, Any, Any]:
    return (
        models.ServiceManifest.__service__,
        models.ServiceManifest,
        Field(
            default=models.ServiceManifest(),
            description="Configuration for the PoWA service, if enabled in site settings.",
        ),
    )


@hookimpl
def get() -> models.ServiceManifest:
    return models.ServiceManifest()


@hookimpl
def rolename(settings: Settings) -> str:
    assert settings.powa
    return settings.powa.role


@hookimpl
def role(settings: Settings, manifest: interface.Instance) -> interface.Role | None:
    name = rolename(settings)
    service_manifest = manifest.service_manifest(models.ServiceManifest)
    return interface.Role(
        name=name, password=service_manifest.password, login=True, superuser=True
    )


@hookimpl
def database(
    settings: Settings, manifest: interface.Instance
) -> interface.Database | None:
    assert settings.powa
    return interface.Database(
        name=settings.powa.dbname,
        extensions=[{"name": item} for item in POWA_EXTENSIONS],
    )



from __future__ import annotations

import pathlib

import pytest

from pglift import exceptions
from pglift.models.system import Instance
from pglift.settings import Settings, _temboard
from pglift.temboard import impl as temboard
from pglift.temboard import models, systemd_unit_templates


@pytest.fixture
def temboard_settings(settings: Settings) -> _temboard.Settings:
    assert settings.temboard is not None
    return settings.temboard


def test_systemd_unit_templates(
    settings: Settings, temboard_execpath: pathlib.Path | None
) -> None:
    assert temboard_execpath
    ((name, content),) = list(systemd_unit_templates(settings=settings, content=True))
    assert name == "pglift-temboard_agent@.service"
    lines = content.splitlines()
    assert (
        f"ExecStart={temboard_execpath} -c {settings.prefix}/etc/temboard-agent/temboard-agent-%i.conf"
        in lines
    )


def test_port(temboard_settings: _temboard.Settings, instance: Instance) -> None:
    try:
        temboard_service = instance.service(models.Service)
    except ValueError:
        temboard_service = None
    if temboard_service:
        port = temboard.port(instance.qualname, temboard_settings)
        assert port == 2345
    else:
        with pytest.raises(exceptions.FileNotFoundError):
            temboard.port(instance.qualname, temboard_settings)

    configpath = pathlib.Path(
        str(temboard_settings.configpath).format(name=instance.qualname)
    )
    original_content = None
    if temboard_service:
        original_content = configpath.read_text()
    else:
        configpath.parent.mkdir(parents=True)  # exists not ok
    try:
        configpath.write_text("[empty section]\n")
        with pytest.raises(LookupError, match="port not found in temboard section"):
            temboard.port(instance.qualname, temboard_settings)
    finally:
        if original_content is not None:
            configpath.write_text(original_content)


def test_password(temboard_settings: _temboard.Settings, instance: Instance) -> None:
    try:
        temboard_service = instance.service(models.Service)
    except ValueError:
        temboard_service = None
    if temboard_service:
        password = temboard.password(instance.qualname, temboard_settings)
        assert password == "dorade"
    else:
        with pytest.raises(exceptions.FileNotFoundError):
            temboard.password(instance.qualname, temboard_settings)

    configpath = pathlib.Path(
        str(temboard_settings.configpath).format(name=instance.qualname)
    )
    original_content = None
    if temboard_service:
        original_content = configpath.read_text()
    else:
        configpath.parent.mkdir(parents=True)  # exists not ok
    try:
        configpath.write_text("[postgresql]\n")
        assert temboard.password(instance.qualname, temboard_settings) is None
    finally:
        if original_content is not None:
            configpath.write_text(original_content)



from __future__ import annotations

import configparser
import getpass
import logging
import shlex
import subprocess
from abc import ABC, abstractmethod, abstractproperty
from collections.abc import Iterator
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

from pglift._compat import Self
from pglift.models.system import PostgreSQLInstance
from pglift.settings import _pgbackrest


@dataclass
class PgbackrestRepoHost(ABC):
    logpath: Path
    port: int
    path: Path
    client_configpath: Path

    logger: logging.Logger

    # Use low timeout values to avoid getting stuck long.
    archive_timeout: int = field(default=2, init=False)
    io_timeout: int = field(default=1, init=False)
    db_timeout: int = field(default=1, init=False)

    @contextmanager
    def edit_config(self) -> Iterator[configparser.ConfigParser]:
        cp = configparser.ConfigParser(strict=True)
        if self.client_configpath.exists():
            with self.client_configpath.open() as f:
                cp.read_file(f)
        yield cp
        with self.client_configpath.open("w") as f:
            cp.write(f)

    @abstractmethod
    def add_stanza(
        self, name: str, instance: PostgreSQLInstance, index: int = 1
    ) -> None:
        ...

    def check_stanza(self, stanza: str) -> None:
        self.run(
            "check",
            "--stanza",
            stanza,
            "--no-archive-check",
            "--io-timeout",
            str(self.io_timeout),
            "--db-timeout",
            str(self.db_timeout),
        )

    def cmd(self, *args: str) -> list[str]:
        """Build a pgbackrest client command."""
        return ["pgbackrest", "--config", str(self.client_configpath)] + list(args)

    def run(self, *args: str) -> subprocess.CompletedProcess[str]:
        """Run a pgbackrest client command from the repository."""
        cmd = self.cmd(*args)
        return self.run_command(cmd)

    def run_command(self, cmd: list[str]) -> subprocess.CompletedProcess[str]:
        self.logger.debug("running: %s", shlex.join(cmd))
        with subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        ) as p:
            assert p.stderr is not None
            stderr = []
            for errline in p.stderr:
                self.logger.debug("%s: %s", cmd[0], errline.rstrip())
                stderr.append(errline)
            assert p.stdout is not None
            stdout = p.stdout.read()
        if p.returncode != 0:
            raise subprocess.CalledProcessError(p.returncode, p.args)
        return subprocess.CompletedProcess(
            p.args, p.returncode, stdout=stdout, stderr="".join(stderr)
        )


@dataclass
class _Server(ABC):
    proc: subprocess.Popen[str] | None = field(default=None, init=False)

    @abstractproperty
    def server_command(self) -> list[str]:
        ...

    def __enter__(self) -> Self:
        assert self.proc is None, "process already started"
        self.proc = subprocess.Popen(self.server_command, text=True)
        self.proc.__enter__()
        return self

    def __exit__(self, *args: Any) -> None:
        assert self.proc is not None, "process not started"
        self.proc.terminate()
        self.proc.__exit__(*args)
        self.proc = None


@dataclass
class PgbackrestRepoHostTLS(PgbackrestRepoHost, _Server):
    server_configpath: Path
    dbhost_cn: str
    ca_file: Path
    repo_cn: str
    repo_certfile: Path
    repo_keyfile: Path

    @property
    def server_command(self) -> list[str]:
        return ["pgbackrest", "server", "--config", str(self.server_configpath)]

    def __post_init__(self) -> None:
        cp = configparser.ConfigParser(strict=True)
        cp.add_section("global")
        cp["global"] = {
            "repo1-path": str(self.path),
            "repo1-retention-full": "2",
            "tls-server-address": "*",
            "tls-server-ca-file": str(self.ca_file),
            "tls-server-cert-file": str(self.repo_certfile),
            "tls-server-key-file": str(self.repo_keyfile),
            "tls-server-auth": f"{self.dbhost_cn}=*",
            "tls-server-port": str(self.port),
            "log-level-console": "off",
            "log-level-file": "detail",
            "log-level-stderr": "info",
            "log-path": str(self.logpath),
        }
        with self.server_configpath.open("w") as f:
            cp.write(f)

        with self.edit_config() as cp:
            cp.add_section("global")
            cp["global"] = {
                "repo1-path": str(self.path),
                "repo1-retention-full": "2",
                "start-fast": "y",
                "archive-timeout": str(self.archive_timeout),
                "log-level-console": "off",
                "log-level-file": "detail",
                "log-level-stderr": "info",
                "log-path": str(self.logpath),
            }

    def add_stanza(
        self, name: str, instance: PostgreSQLInstance, index: int = 1
    ) -> None:
        settings = instance._settings
        pgbackrest_settings = settings.pgbackrest
        assert pgbackrest_settings is not None
        host_config_path = pgbackrest_settings.configpath
        assert isinstance(pgbackrest_settings.repository, _pgbackrest.TLSHostRepository)
        host_port = pgbackrest_settings.repository.port
        user = settings.postgresql.backuprole.name
        socket_path = settings.postgresql.socket_directory
        pg = f"pg{index}"
        with self.edit_config() as cp:
            if not cp.has_section(name):
                cp.add_section(name)
            cp[name].update(
                {
                    f"{pg}-host": self.dbhost_cn,
                    f"{pg}-host-port": str(host_port),
                    f"{pg}-host-type": "tls",
                    f"{pg}-host-config-path": str(host_config_path),
                    f"{pg}-host-ca-file": str(self.ca_file),
                    f"{pg}-host-cert-file": str(self.repo_certfile),
                    f"{pg}-host-key-file": str(self.repo_keyfile),
                    f"{pg}-path": str(instance.datadir),
                    f"{pg}-port": str(instance.port),
                    f"{pg}-user": user,
                    f"{pg}-socket-path": str(socket_path),
                }
            )
        self.run(
            "server-ping",
            "--io-timeout",
            str(self.io_timeout),
            "--tls-server-address",
            self.dbhost_cn,
            "--tls-server-port",
            str(host_port),
        )
        self.run("stanza-create", "--stanza", name, "--no-online")
        self.run("verify", "--stanza", name)
        self.run("repo-ls")
        self.check_stanza(name)


@dataclass
class ServerSSH(_Server):
    ssh_path: Path
    host_keyfile: Path
    port: int

    @property
    def server_command(self) -> list[str]:
        return [
            "/usr/sbin/sshd",
            "-D",
            "-e",
            "-o",
            f"AuthorizedKeysFile={self.ssh_path / 'authorized_keys'}",
            "-o",
            "AuthenticationMethods=publickey",
            "-o",
            "PidFile=no",
            "-o",
            "StrictModes=no",
            "-h",
            str(self.host_keyfile),
            "-p",
            str(self.port),
        ]


@dataclass
class PgbackrestRepoHostSSH(PgbackrestRepoHost, ServerSSH):
    host_keyfile: Path
    user_keyfile: Path
    cmd_ssh: Path
    dbhost_host: str
    dbhost_port: int
    dbhost_user_keyfile: Path
    dbhost_cmd_ssh: Path

    def __post_init__(self) -> None:
        cp = configparser.ConfigParser(strict=True)

        current_user = getpass.getuser()
        with self.edit_config() as cp:
            cp.add_section("global")
            cp["global"] = {
                "repo1-path": str(self.path),
                "repo1-retention-full": "2",
                "repo1-host-user": current_user,
                "start-fast": "y",
                "archive-timeout": str(self.archive_timeout),
                "log-level-console": "info",
                "log-level-file": "detail",
                "log-level-stderr": "info",
                "log-path": str(self.logpath),
                "cmd-ssh": str(self.cmd_ssh),
            }

    def add_stanza(
        self, name: str, instance: PostgreSQLInstance, index: int = 1
    ) -> None:
        settings = instance._settings
        pgbackrest_settings = settings.pgbackrest
        assert pgbackrest_settings is not None
        host_config_path = pgbackrest_settings.configpath
        assert isinstance(pgbackrest_settings.repository, _pgbackrest.SSHHostRepository)
        socket_path = settings.postgresql.socket_directory
        pg = f"pg{index}"
        current_user = getpass.getuser()
        user = settings.postgresql.backuprole.name
        with self.edit_config() as cp:
            if not cp.has_section(name):
                cp.add_section(name)
            cp[name].update(
                {
                    f"{pg}-host": self.dbhost_host,
                    f"{pg}-host-port": str(self.dbhost_port),
                    f"{pg}-host-type": "ssh",
                    f"{pg}-host-user": current_user,
                    f"{pg}-host-config-path": str(host_config_path),
                    f"{pg}-path": str(instance.datadir),
                    f"{pg}-port": str(instance.port),
                    f"{pg}-user": user,
                    f"{pg}-socket-path": str(socket_path),
                }
            )
        self.ping_ssh(
            f"{current_user}@{self.dbhost_host}",
            "-p",
            str(self.dbhost_port),
            "-i",
            str(self.user_keyfile),
            "-o",
            "StrictHostKeyChecking=no",
        )
        self.ping_ssh(
            f"{current_user}@{self.dbhost_host}",
            "-p",
            str(self.port),
            "-i",
            str(self.dbhost_user_keyfile),
            "-o",
            "StrictHostKeyChecking=no",
        )
        self.run(
            "stanza-create",
            "--stanza",
            name,
            "--no-online",
            "--log-level-console=debug",
        )
        self.run("verify", "--stanza", name)
        self.run("repo-ls")
        self.check_stanza(name)

    def ping_ssh(self, *args: str) -> subprocess.CompletedProcess[str]:
        """Check SSH connection from/to the repository."""
        cmd = ["ssh"] + list(args)
        return self.run_command(cmd)



from __future__ import annotations

import logging

import psycopg.rows
from psycopg import sql

from . import db
from .models import interface

logger = logging.getLogger(__name__)


def ls(cnx: db.Connection) -> list[interface.Schema]:
    """Return list of schemas of database."""
    with cnx.cursor(row_factory=psycopg.rows.class_row(interface.Schema)) as cur:
        cur.execute(db.query("list_schemas"))
        return cur.fetchall()


def current_role(cnx: db.Connection) -> str:
    with cnx.cursor(row_factory=psycopg.rows.args_row(str)) as cur:
        r = cur.execute("SELECT CURRENT_ROLE").fetchone()
        assert r is not None
        return r


def alter_owner(cnx: db.Connection, name: str, owner: str) -> None:
    opts = sql.SQL("OWNER TO {}").format(sql.Identifier(owner))
    logger.info("setting '%s' schema owner to '%s'", name, owner)
    cnx.execute(
        db.query("alter_schema", schema=psycopg.sql.Identifier(name), opts=opts)
    )


def apply(cnx: db.Connection, schema: interface.Schema, dbname: str) -> bool:
    """Apply the state defined by 'schema' in connected database and return
    True if something changed.
    """
    for existing in ls(cnx):
        if schema.name == existing.name:
            if schema.state is interface.PresenceState.absent:
                logger.info("dropping schema %s from database %s", schema.name, dbname)
                cnx.execute(
                    db.query("drop_schema", schema=psycopg.sql.Identifier(schema.name))
                )
                return True

            new_owner = schema.owner or current_role(cnx)
            if new_owner != existing.owner:
                alter_owner(cnx, schema.name, new_owner)
                return True
            return False

    if schema.state is not interface.PresenceState.absent:
        create(cnx, schema, dbname)
        return True
    return False


def create(cnx: db.Connection, schema: interface.Schema, dbname: str) -> None:
    msg, args = "creating schema '%(name)s' in database %(dbname)s", {
        "name": schema.name,
        "dbname": dbname,
    }
    opts = []
    if schema.owner is not None:
        opts.append(sql.SQL("AUTHORIZATION {}").format(sql.Identifier(schema.owner)))
        msg += " with owner '%(owner)s'"
        args["owner"] = schema.owner

    logger.info(msg, args)
    cnx.execute(
        db.query(
            "create_schema",
            schema=psycopg.sql.Identifier(schema.name),
            options=sql.SQL(" ").join(opts),
        )
    )



from __future__ import annotations

from datetime import datetime
from functools import partial
from typing import Any

import attrs
import click

from .. import cmd, postgresql, types
from ..cli.instance import instance_identifier
from ..cli.util import (
    Command,
    OutputFormat,
    instance_identifier_option,
    output_format_option,
    pass_component_settings,
    pass_ctx,
    print_json_for,
    print_table_for,
)
from ..ctx import Context
from ..models import system
from ..settings import _pgbackrest
from . import base, models, repo_path

pass_pgbackrest_settings = partial(pass_component_settings, base, "pgbackrest")


@click.command(
    "pgbackrest",
    hidden=True,
    cls=Command,
    context_settings={"ignore_unknown_options": True},
)
@instance_identifier_option
@click.argument("command", nargs=-1, type=click.UNPROCESSED)
@pass_pgbackrest_settings
@click.pass_context
def pgbackrest(
    context: click.Context,
    /,
    settings: _pgbackrest.Settings,
    command: tuple[str, ...],
    **kwargs: Any,
) -> None:
    """Proxy to pgbackrest operations on an instance"""
    s = context.obj.instance.service(models.Service)
    cmd_args = base.make_cmd(s.stanza, settings, *command)
    cmd.run(cmd_args, capture_output=False, check=True)


@click.command("backup", cls=Command)
@instance_identifier(nargs=1)
@click.option(
    "--type",
    "backup_type",
    type=click.Choice([t.name for t in types.BackupType]),
    default=types.BackupType.default().name,
    help="Backup type",
    callback=lambda ctx, param, value: types.BackupType(value),
)
@pass_pgbackrest_settings
@pass_ctx
def instance_backup(
    ctx: Context,
    settings: _pgbackrest.Settings,
    instance: system.Instance,
    backup_type: types.BackupType,
) -> None:
    """Back up PostgreSQL INSTANCE"""
    with ctx.lock:
        repo_path.backup(instance, settings, type=backup_type)


@click.command("restore", cls=Command)
@instance_identifier(nargs=1)
@click.option("--label", help="Label of backup to restore")
@click.option("--date", type=click.DateTime(), help="Date of backup to restore")
@pass_pgbackrest_settings
@pass_ctx
def instance_restore(
    ctx: Context,
    settings: _pgbackrest.Settings,
    instance: system.Instance,
    label: str | None,
    date: datetime | None,
) -> None:
    """Restore PostgreSQL INSTANCE"""
    postgresql.check_status(instance, types.Status.not_running)
    if label is not None and date is not None:
        raise click.BadArgumentUsage(
            "--label and --date arguments are mutually exclusive"
        ) from None
    with ctx.lock:
        base.restore(ctx, instance, settings, label=label, date=date)


@click.command("backups", cls=Command)
@output_format_option
@instance_identifier(nargs=1)
@pass_pgbackrest_settings
def instance_backups(
    settings: _pgbackrest.Settings,
    instance: system.Instance,
    output_format: OutputFormat,
) -> None:
    """List available backups for INSTANCE"""
    values = list(map(attrs.asdict, base.iter_backups(instance, settings)))
    if output_format == OutputFormat.json:
        print_json_for(values)
    else:
        print_table_for(values, title=f"Available backups for instance {instance}")
